"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.handleServeStaticContent = void 0;
const fs_1 = __importDefault(require("fs"));
const path_1 = __importDefault(require("path"));
const app_root_path_1 = __importDefault(require("app-root-path"));
const __1 = require("..");
const providers_1 = require("../services/common/providers");
const auth_middleware_1 = require("./auth/auth.middleware");
const user_model_1 = require("./auth/user.model");
const middleware_api_1 = require("./middleware.api");
// Configure this as needed
const RELATIVE_ROOT_BUILD_PATH = "client/build";
const CLIENT_BUILD_PATH = path_1.default.join(`${app_root_path_1.default}`, RELATIVE_ROOT_BUILD_PATH);
const handleServeStaticContent = (app, swagger, customServeStaticMiddleware) => {
    middleware_api_1.logMiddlewareStart(`Serve Static Content`);
    middleware_api_1.middlewareLogger.debug(`Serving from: ${CLIENT_BUILD_PATH}`);
    // If swagger is enabled, we need to serve the redirect page for swagger when the redirect url is hit
    if (swagger) {
        app.get("/oauth2-redirect.html*", auth_middleware_1.ensureSessionAuthRedirectUrl, (req, res) => {
            res.set("Content-Security-Policy", "default-src https: 'unsafe-eval' 'unsafe-inline'; object-src 'none'");
            res.sendFile(`${app_root_path_1.default}/node_modules/swagger-ui-dist/oauth2-redirect.html`);
        });
    }
    const CustomExpressMiddleware = customServeStaticMiddleware
        ? customServeStaticMiddleware
        : []; // if undefined or null assigned to empty array so spread below works
    app.get("*", handleClientFileRequest, auth_middleware_1.ensureSessionAuthRedirectUrl, handleMaintenanceMode, ...CustomExpressMiddleware, (req, res) => __awaiter(void 0, void 0, void 0, function* () {
        // Proxy all other requests to the frontend
        // Default with session auth guard to allow client to redirect back after login
        // in IES the default deployment of static files is through node which is why we don't ignore the proxy in production
        middleware_api_1.middlewareLogger.debug(`Maintenance mode not enabled for this request`);
        middleware_api_1.middlewareLogger.info(`Serving index.html for user`);
        res.sendFile(`${CLIENT_BUILD_PATH}/index.html`);
    }));
};
exports.handleServeStaticContent = handleServeStaticContent;
const serveIndexOrMaintenancePage = (req, res) => __awaiter(void 0, void 0, void 0, function* () {
    const isUnderMaintenance = yield fetchMaintenanceStatus();
    if (isUnderMaintenance) {
        serveMaintenancePage(req, res);
        return;
    }
    serveIndexPage(req, res);
});
const fetchMaintenanceStatus = () => __awaiter(void 0, void 0, void 0, function* () {
    var _a;
    const config = __1.ApplicationConfig.get();
    const provider = new providers_1.CommonServicesProvider({});
    try {
        const { underMaintenance } = yield provider.getServiceById((_a = config.common) === null || _a === void 0 ? void 0 : _a.serviceId);
        middleware_api_1.middlewareLogger.debug(`Service data retrieved, maintenance mode is set to ${underMaintenance}`);
        return underMaintenance;
    }
    catch (e) {
        middleware_api_1.middlewareLogger.error(`Failed to fetch service data: ${e}`);
        return false;
    }
});
const serveMaintenancePage = (req, res) => {
    middleware_api_1.middlewareLogger.info(`Serving maintenance.html to client`);
    const path = `${CLIENT_BUILD_PATH}/maintenance.html`;
    fs_1.default.stat(path, (err, stat) => {
        if (err)
            return res.status(404).send();
        if (stat.isDirectory())
            return res.status(404).send();
        return res.sendFile(path);
    });
};
const serveIndexPage = (req, res) => {
    middleware_api_1.middlewareLogger.info(`Serving index.html to client`);
    const path = `${CLIENT_BUILD_PATH}/index.html`;
    fs_1.default.stat(path, (err, stat) => {
        if (err)
            return res.status(404).send();
        if (stat.isDirectory())
            return res.status(404).send();
        return res.sendFile(path);
    });
};
/**
 * Serves a file client requested if it exists, if not continues
 */
const handleClientFileRequest = (req, res, next) => __awaiter(void 0, void 0, void 0, function* () {
    const path = `${CLIENT_BUILD_PATH}${req.path}`;
    middleware_api_1.middlewareLogger.debug(`Servicing file request at ${path}`);
    fs_1.default.stat(path, (err, stat) => {
        if (err)
            return next();
        if (stat.isDirectory())
            return next();
        return res.sendFile(path);
    });
});
/**
 * If server is running in maintenance mode, will call service to determine application status
 * Admin users see the index.html page, all other users see either index.html or maintenance.html
 * depending on maintenance status
 */
const handleMaintenanceMode = (req, res, next) => __awaiter(void 0, void 0, void 0, function* () {
    const config = __1.ApplicationConfig.get();
    if (config.app.enableMaintenanceMode) {
        if (!req.isAuthenticated()) {
            middleware_api_1.middlewareLogger.debug(`User is not authenticated, determining index or maintenance page to serve`);
            yield serveIndexOrMaintenancePage(req, res);
            return;
        }
        const user = new user_model_1.AuthenticatedUser(req.user);
        const isAdmin = user.isAdmin;
        if (isAdmin) {
            middleware_api_1.middlewareLogger.debug("Admin user: will serve index.html regardless of maintenance status");
            yield serveIndexPage(req, res);
            return;
        }
        middleware_api_1.middlewareLogger.debug("Standard user: determining index of maintenance page to serve");
        yield serveIndexOrMaintenancePage(req, res);
        return;
    }
    next();
});
