"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.configureAppHealth = void 0;
const terminus_1 = require("@godaddy/terminus");
const common_1 = require("./common");
const config_1 = require("./config");
/**
 * Optionally include a value to be included in
 * the status info in the health check response
 */
const healthCheck = () => Promise.resolve();
/**
 * When Kubernetes or a user deletes a Pod, Kubernetes will notify it and wait for gracePeriod seconds before killing it.
 * During that time window (30 seconds by default), the Pod is in the terminating state and will be removed from any Services by a controller.
 *
 * The Pod itself needs to catch the SIGTERM signal and start failing any readiness probes. During this time, it is possible that load-balancers
 * (like the nginx ingress controller) don't remove the Pods "in time", and when the Pod dies, it kills live connections.
 *
 * To make sure you don't lose any connections, it is recommended to delay the shutdown with the number of milliseconds that's defined by the
 * readiness probe in your deployment configuration. To help with this, terminus exposes the beforeShutdown hook.
 */
const beforeShutdown = () => {
    common_1.logger.log("Server is shutting down...");
    return new Promise((resolve) => {
        setTimeout(resolve, config_1.baseConfig.health.beforeShutdownTimeout);
    });
};
const onSignal = () => {
    common_1.logger.log("Server is starting cleanup");
    return Promise.all([
    // Place clean logic such as closing database connections
    ]);
};
const onShutdown = () => {
    common_1.logger.log("Cleanup finished, server is shutting down");
    return Promise.resolve();
};
const terminusOptions = {
    healthChecks: {
        [config_1.baseConfig.health.checkUrl]: healthCheck,
        verbatim: true, // [optional = false] use object returned from /healthcheck verbatim in response
    },
    signal: "SIGINT",
    beforeShutdown,
    onSignal,
    onShutdown, // [optional] called right before exiting
};
/**
 * Handles Health Checks and graceful termination signals
 * for the server
 * @param server the http server
 */
const configureAppHealth = (server) => {
    terminus_1.createTerminus(server, terminusOptions);
};
exports.configureAppHealth = configureAppHealth;
