"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.requirePolicy = exports.ensureBasicAuth = exports.ensureAdminUser = exports.ensureAuthUser = exports.ensureSessionAuthRedirectUrl = exports.ensureAuthenticated = void 0;
const url_1 = require("url");
const node_fetch_1 = __importDefault(require("node-fetch"));
const common_1 = require("../../common");
const env_config_1 = require("../../config/env.config");
const utils_1 = require("../../utils");
const middleware_api_1 = require("./../middleware.api");
const policies_service_1 = require("./policies.service");
const routes_1 = require("./routes");
/**
 * Router-level middleware used to ensure a request has been authenticated.
 *
 * If request has not been authentic, request will be redirected to AUTH_ROUTES.LOGIN.
 *
 * @type {void}
 * @param {Request} req
 * @param {Response} res
 * @param {NextFunction} next
 */
/**
 * Helper function to perform OpenID Connect introspection by making a call
 * to w3id instropection endpoint passing in access token along with the app's
 * client id and client secret. If the response contains a property named 'active'
 * that is set to true, then access token is considered valid.
 *
 * @type {string}
 * @param accesstoken
 */
const introspectAccesstoken = (accesstoken) => __awaiter(void 0, void 0, void 0, function* () {
    const params = new url_1.URLSearchParams({
        token: accesstoken,
        client_id: env_config_1.baseConfig.openid.client_id,
        client_secret: env_config_1.baseConfig.openid.client_secret,
    });
    middleware_api_1.middlewareLogger.info(`Submitting introspection request.`);
    return node_fetch_1.default(env_config_1.baseConfig.openid.introspection_url, {
        method: "post",
        body: params,
    })
        .then((res) => res.json())
        .then((json) => {
        return json;
    });
});
/**
 * The w3id access token is composed of the username and password and
 * is connected to the client_id and client_secret of the app that
 * requested the user to login.
 *
 * The token is parsed out from the authorization header in this expected format:
 * "authorization: Bearer accessToken"
 *
 * @type {string}
 * @param req
 */
const getAccessToken = (req) => {
    try {
        return req.get("authorization").split(" ")[1];
    }
    catch (e) {
        throw new common_1.HTTP401Error();
    }
};
/**
 *
 * Helper function to determine not-authenticated response based on
 * whether the template is being used as a backend-for-frontend or as
 * a pure backend service.
 *
 * @param req
 * @param res
 */
const handleNotAuthenticatedResponse = (req, res) => {
    if (env_config_1.baseConfig.app.serviceType === "BFF") {
        middleware_api_1.middlewareLogger.info(`Request lacks authorization header. Setting session url for redirection.`);
        const reqSession = req;
        reqSession.session.originalUrl = req.originalUrl;
        middleware_api_1.middlewareLogger.debug(`Redirecting to ${routes_1.AUTH_ROUTES.LOGIN} ...`);
        res.redirect(routes_1.AUTH_ROUTES.LOGIN);
    }
    else {
        middleware_api_1.middlewareLogger.info(`Request lacks authorization header. Throwing unauthorized error.`);
        throw new common_1.HTTP401Error();
    }
};
/**
 * Router-level middleware used to ensure a request has been authenticated.
 *
 * If the request has not been authenticated and if the request contains an
 * authorization header, then it implies that this template is being used as
 * a pure backend service and the access token needs to be verified via
 * introspection.
 *
 * Next can be optionally passed as an argument, allowing it to be called outside
 * of this middleware file for manual authentication elsewhere.
 *
 * @type {void}
 * @param {Request} req
 * @param {Response} res
 * @param {NextFunction} next
 */
const ensureAuthenticated = (req, res, next) => __awaiter(void 0, void 0, void 0, function* () {
    if (!req.isAuthenticated()) {
        middleware_api_1.middlewareLogger.info(`User failed authentication against protected route: ${req.url}.`);
        if (req.get("Authorization")) {
            const accessToken = getAccessToken(req);
            const introspectionResponse = yield introspectAccesstoken(accessToken);
            if (introspectionResponse.active) {
                req.user = {
                    id: introspectionResponse === null || introspectionResponse === void 0 ? void 0 : introspectionResponse.emailAddress,
                    _json: introspectionResponse,
                };
                return next ? next() : true;
            }
            else {
                middleware_api_1.middlewareLogger.info(`User's access token is expired or not valid.`);
                throw new common_1.HTTP401Error();
            }
        }
        else {
            handleNotAuthenticatedResponse(req, res);
        }
    }
    else {
        return next ? next() : true;
    }
});
exports.ensureAuthenticated = ensureAuthenticated;
/**
 * Router-level middleware used to ensure a request has been authenticated.
 *
 * If the request has not been authenticated, request will be redirected to the requests originalUrl
 *
 * @type {void}
 * @param {Request} req
 * @param {Response} res
 * @param {NextFunction} next
 */
const ensureSessionAuthRedirectUrl = (req, res, next) => {
    if (!req.isAuthenticated()) {
        middleware_api_1.middlewareLogger.info(`User failed authentication against protected route: ${req.url}. Setting session url ...`);
        const reqSession = req;
        reqSession.session.originalUrl = req.originalUrl;
    }
    return next();
};
exports.ensureSessionAuthRedirectUrl = ensureSessionAuthRedirectUrl;
/**
 * Router-level middleware used to ensure a request has been authenticated.
 *
 * If the request has not been authenticated, HTTP401Error will be thrown.
 *
 * @type {void}
 * @param {Request} req
 * @param {Response} res
 * @param {NextFunction} next
 */
const ensureAuthUser = (req, res, next) => {
    if (!req.isAuthenticated()) {
        middleware_api_1.middlewareLogger.info(`User failed authentication against protected endpoint: ${req.url} `);
        throw new common_1.HTTP401Error();
    }
    return next();
};
exports.ensureAuthUser = ensureAuthUser;
/**
 * Router-level middleware used to validate if a requesting user has the configured admin blue group
 *
 * If the requesting user does not have the configured admin blue group, HTTP403Error will be thrown.
 *
 * @type {void}
 * @param {Request} req
 * @param {Response} res
 * @param {NextFunction} next
 */
const ensureAdminUser = (req, res, next) => {
    const user = req.user; // extends the express user modeling, allowing _json.blueGroups
    if (!utils_1.isAdministrator(user._json.blueGroups)) {
        middleware_api_1.middlewareLogger.error(`User not authorized to access this API`);
        throw new common_1.HTTP403Error();
    }
    else {
        next();
    }
};
exports.ensureAdminUser = ensureAdminUser;
/**
 * Router-level middleware used to validate a request against the basic auth credentials stored in the server config
 *
 * If credentials are not supplied, or are incorrect, HTTP403Error will be thrown. If a developer doesn't provide basicAuth
 * in the config, an error will get thrown. The desired behavior in this case is to have the server return 500.
 *
 * @type {void}
 * @param {Request} req
 * @param {Response} res
 * @param {NextFunction} next
 */
const ensureBasicAuth = (req, res, next) => {
    var _a, _b;
    // Developer errors - Basic Auth credentials are not stored in the config
    if (!((_a = env_config_1.ApplicationConfig.get().basicAuth) === null || _a === void 0 ? void 0 : _a.username) ||
        !((_b = env_config_1.ApplicationConfig.get().basicAuth) === null || _b === void 0 ? void 0 : _b.pass)) {
        middleware_api_1.middlewareLogger.error("Application config lacks basicAuth credentials to compare against");
        throw new Error("BasicAuth not set");
    }
    const userAuthHeader = req.get("Authorization");
    if (!userAuthHeader) {
        middleware_api_1.middlewareLogger.info(`User failed authentication against ${req.url}: No authorization was provided`);
        throw new common_1.HTTP403Error("No authorization provided");
    }
    const headersMatch = userAuthHeader ===
        "Basic " +
            Buffer.from(`${env_config_1.ApplicationConfig.get().basicAuth.username}:${env_config_1.ApplicationConfig.get().basicAuth.pass}`).toString("base64");
    if (!headersMatch) {
        middleware_api_1.middlewareLogger.info(`User failed authentication against ${req.url}: Invalid credentials provided`);
        throw new common_1.HTTP403Error("Invalid credentials provided");
    }
    return next();
};
exports.ensureBasicAuth = ensureBasicAuth;
/**
 * Router-level middleware used to ensure a user has accepted the applications global policy
 *
 * If the request does not have a user property, the request will be redirected to AUTH_ROUTES.LOGIN allow the user to login.
 *
 * If the user making the request has not accepted the applications policy, validateUserWithServiceIdApi, the request will be redirected to the req.requestUrl
 *
 * @type {void}
 * @param {Request} req
 * @param {Response} res
 * @param {NextFunction} next
 */
const requirePolicy = (req, res, next) => __awaiter(void 0, void 0, void 0, function* () {
    var _a;
    if (req.user) {
        const config = env_config_1.ApplicationConfig.get();
        if ((_a = config.common) === null || _a === void 0 ? void 0 : _a.policiesApi) {
            const authUser = req.user;
            const user = yield policies_service_1.validateUserWithServiceIdApi(authUser.id, config);
            if (user.accept) {
                middleware_api_1.middlewareLogger.info(`${authUser._json.emailAddress} accepted policy!`);
                return next();
            }
            middleware_api_1.middlewareLogger.info("User did not accept policy!");
            middleware_api_1.middlewareLogger.debug(`Redirecting to ${user.redirectUrl} ...`);
            res.redirect(user.redirectUrl);
        }
        else {
            middleware_api_1.middlewareLogger.info("Application has no policy configured");
            return next();
        }
    }
    else {
        middleware_api_1.middlewareLogger.info("User is no longer authenticated when attempting to check policy!");
        middleware_api_1.middlewareLogger.debug(`Redirecting to ${routes_1.AUTH_ROUTES.LOGIN} ...`);
        res.redirect(routes_1.AUTH_ROUTES.LOGIN);
    }
});
exports.requirePolicy = requirePolicy;
