import { getBooleanEnvVar, getIntegerEnvVar, getStringEnvVar, } from "../utils/env.utils";
/**
 * Set default NODE ENV variables
 */
// Allows for certs to work with HTTPS locally
process.env.NODE_TLS_REJECT_UNAUTHORIZED = "0";
const environment = getStringEnvVar("NODE_ENV", "development");
export const isProduction = environment === "production";
export const isDevelopment = environment === "development";
export const isTest = environment === "test"; // Defines that we are running tests, but locally
export const isRemoteTest = environment === "remote-test"; // Defines that we are running tests, but externally (like in a deployment pipeline)
// Set test specific environments for testsuite
// TODO find away to include this elsewhere
if (isTest || isRemoteTest) {
    process.env.BLUEGROUPS_ADMIN = "ADMIN";
    process.env.BLUEGROUPS_AUDITOR = "AUDITOR";
    process.env.BLUEGROUPS_APP_USER = "NODE_TEMPLATE";
    process.env.BLUEGROUPS_DEVOPS = "DEVOPS";
    process.env.BLUEGROUPS_PRODUCTOWNER = "PRODUCT_OWNER";
    process.env.BLUEGROUPS_EXTERNAL_USER = "EXTERNAL";
    process.env.BLUEGROUPS_TEAM_MEMBER = "TEAM_MEMBER";
    process.env.LOG_LEVEL = "silent";
    process.env.DASHBOARD_SERVICE_API =
        "https://ies-dashboard-service.dev.identity-services.intranet.ibm.com/api/v1";
    process.env.OFFERING_ID = "6102eaa24908dd09cddfee39";
}
if (isRemoteTest) {
    process.env.HSK_API =
        "https://ies-hsk-management.dev.identity-services.intranet.ibm.com/api/v1";
    process.env.NODE_TLS_REJECT_UNAUTHORIZED = "0";
    process.env.OPENID_CALLBACK_URL = "default_callback_url";
    process.env.OPENID_CLIENT_ID = "default_client_id";
    process.env.OPENID_CLIENT_SECRET = "default_client_secret";
    process.env.OPENID_DISCOVERY_URL = "default_discovery_url";
    process.env.OPENID_LOGOUT_URL = "default_logout_url";
    process.env.POLICIES_API = "default_policies_api";
    process.env.BASICAUTH_PASS = "default_BASICAUTH_PASS";
    process.env.SERVICE_ID = "ies-hsk-registration";
    process.env.BASICAUTH_USERNAME = "default_BASICAUTH_USERNAME";
    process.env.PORT = "8443";
    process.env.REDIS_CERT_FILE = "redis.cer";
    process.env.REDIS_PORT = "31440";
    process.env.REDIS_PREFIX = "ies-hsk-registration-frontend:";
    process.env.REDIS_URL = "redis_url";
    process.env.SERVE_STATIC = "true";
    process.env.UNAUTHORIZED_REDIRECT_URL = "default_unauthorized_redirect_url";
    process.env.W3_SUPPORT_PAGE = "https://w3.ibm.com/help/#/article/hsk";
}
class Config {
    constructor() {
        this.baseConfig = Object.freeze({
            app: {
                serviceId: getStringEnvVar("APP_NAME", "ies-node-api-template"),
                serviceType: getStringEnvVar("SERVICE_TYPE", "BFF"),
                apiBasePath: getStringEnvVar("API_BASE_PATH", "/api"),
                environment,
                port: getIntegerEnvVar("PORT", 8443),
                serveStatic: getBooleanEnvVar("SERVE_STATIC", false),
                enableMaintenanceMode: getBooleanEnvVar("ENABLE_MAINTENANCE_MODE", false),
                swagger: getBooleanEnvVar("SWAGGER", false),
                logLevel: getStringEnvVar("LOG_LEVEL", "info"),
                unauthorizedRedirectUrl: getStringEnvVar("UNAUTHORIZED_REDIRECT_URL", "https://ies-portal.prod.identity-services.intranet.ibm.com/#/"),
                enableHttpRequestLogging: getBooleanEnvVar("ENABLE_HTTP_REQUEST_LOGGING", true),
                useHttps: getBooleanEnvVar("USE_HTTPS", true),
                requireLogin: getBooleanEnvVar("REQUIRE_LOGIN", true), // Boolean that determines whether or not handleLogin middleware is automatically included in auth middleware (OIDC flow)
            },
            health: {
                checkUrl: getStringEnvVar("HEALTH_CHECK_URL", "/actuator/health"),
                beforeShutdownTimeout: getIntegerEnvVar("TIMEOUT_BEFORE_SHUTDOWN", 6000), // modifies terminus behavior, see health.ts file for more information
            },
            session: {
                secret: getStringEnvVar("SESSION_SECRET", "nodeapi"), // This is a required option for the secret to sign the session ID cookie, it can be a string or an array of multiple string secrets.. see session.middleware
            },
            redis: {
                // EFFECTS TEMPLATE session middleware - see session.middleware.ts
                url: getStringEnvVar("REDIS_URL", ""),
                port: getIntegerEnvVar("REDIS_PORT", 6379),
                cert: getStringEnvVar("REDIS_CERT_FILE", "redis.cer"),
                prefix: getStringEnvVar("REDIS_PREFIX", "ies-node-api-template"), // sets the keyPrefix option for redis, see RedisClient
            },
            mongodb: {
                url: getStringEnvVar("MONGODB_URL", ""), // url for mongodb, if used
            },
            // Config variables related to the openid (oidc) login flow - strategy is defined in OpenIdConnect strategy as scope "openid" response_type "code"
            // Additional certificates may be required to make this login flow work, see certs defined in the strategy
            // In order to make oidc flow work, the variables here that are consumed by the strategy must be present and correct. The other oidc variables are used for other purposes
            // Personally, I found this helpful in understanding our login flow - https://darutk.medium.com/diagrams-of-all-the-openid-connect-flows-6968e3990660#:~:text=1.%20response_type%3Dcode
            openid: {
                client_id: getStringEnvVar("OPENID_CLIENT_ID", "_"),
                client_secret: getStringEnvVar("OPENID_CLIENT_SECRET", "_"),
                discovery_url: getStringEnvVar("OPENID_DISCOVERY_URL", "https://preprod.login.w3.ibm.com/oidc/endpoint/default/.well-known/openid-configuration"),
                callback_url: getStringEnvVar("OPENID_CALLBACK_URL", "https://localhost:8443/api/openid_connect_login"),
                logout_url: getStringEnvVar("OPENID_LOGOUT_URL", "https://w3id.alpha.sso.ibm.com/pkmslogout"),
                introspection_url: getStringEnvVar("OPENID_INTROSPECTION_URL", "https://preprod.login.w3.ibm.com/oidc/endpoint/default/introspect"),
                token_url: getStringEnvVar("OPENID_TOKEN_URL", "https://preprod.login.w3.ibm.com/oidc/endpoint/default/token"),
                // not consumed by the strategy, as our strategy (I believe) abstracts this away for us via use of the discovery_url (or callback_url?), but if implementing this flow without using the strategy
                // (as is the case with Swagger), we need to define the endpoint to send our code to (once retrieved from the authorization server) so that we can receive a token back
            },
            bluegroups: {
                // bluegroups is defined on the return data from w3id, we use it to construct properties for our user model
                // e.g. if a user has reg_admin as a value in their bluegroups array and admin bluegroups is set to reg_admin
                // that user would have the isAdmin boolean set to true on their user model
                admin: [getStringEnvVar("BLUEGROUPS_ADMIN", "")],
                auditor: [getStringEnvVar("BLUEGROUPS_AUDITOR", "")],
                appUser: [getStringEnvVar("BLUEGROUPS_APP_USER", "")],
                bypassAppUserAuth: getBooleanEnvVar("BYPASS_APP_USER_AUTH", false),
                devOps: [getStringEnvVar("BLUEGROUPS_DEVOPS", "")],
                po: [getStringEnvVar("BLUEGROUPS_PRODUCTOWNER", "")],
                external: [getStringEnvVar("BLUEGROUPS_EXTERNAL_USER", "")],
                teamMember: [getStringEnvVar("BLUEGROUPS_TEAM_MEMBER", "")],
                helpDesk: [getStringEnvVar("BLUEGROUPS_HELP_DESK", "")], // defines IT Helpdesk permissions for the application
            },
        });
        // Anything not listed under the baseConfig gets added to the customConfig optionally by the user
        // This customConfig is needed to make certain features work with the template
        this.customConfig = {};
    }
    // adds 'api' prop to 'customConfig' object and spreads the values of the provided 'apis' object,
    useServices(apis) {
        this.customConfig = Object.assign(Object.assign({}, this.customConfig), { api: Object.assign({}, apis) });
        return this;
    }
    // adds values to the 'bluegroups' prop of the 'customConfig' object - spreads the values of
    // the provided 'roles' object
    addBluegroups(roles) {
        const baseConfigBlueGroups = Reflect.ownKeys(this.get().bluegroups);
        const newBluegroups = Reflect.ownKeys(roles).reduce((acc, newBlueGroupKey) => {
            if (!baseConfigBlueGroups.includes(newBlueGroupKey)) {
                return Object.assign(Object.assign({}, acc), { [newBlueGroupKey]: roles[newBlueGroupKey] });
            }
            else {
                const newGroups = Array.isArray(roles[newBlueGroupKey])
                    ? roles[newBlueGroupKey]
                    : [roles[newBlueGroupKey]];
                return Object.assign(Object.assign({}, acc), { [newBlueGroupKey]: [
                        ...this.get().bluegroups[newBlueGroupKey],
                        ...newGroups,
                    ] });
            }
        }, {});
        this.customConfig = Object.assign(Object.assign({}, this.customConfig), { bluegroups: Object.assign(Object.assign({}, this.baseConfig.bluegroups), newBluegroups) });
        return this;
    }
    // EFFECTS TEMPLATE MIDDLEWARE
    usePolicies(api, username, pwd, serviceId) {
        var _a;
        this.customConfig = Object.assign(Object.assign({}, this.customConfig), { common: Object.assign(Object.assign({}, (_a = this.customConfig) === null || _a === void 0 ? void 0 : _a.common), { policiesApi: api, serviceId: serviceId }) });
        this.setBasicAuth(username, pwd);
        return this;
    }
    // EFFECTS TEMPLATE MIDDLEWARE
    enableMaintenanceMode(api, username, password, serviceId) {
        var _a;
        this.customConfig = Object.assign(Object.assign({}, this.customConfig), { common: Object.assign(Object.assign({}, (_a = this.customConfig) === null || _a === void 0 ? void 0 : _a.common), { servicesApi: api, serviceId }) });
        this.setBasicAuth(username, password);
        return this;
    }
    // EFFECTS TEMPLATE MIDDLEWARE
    setBasicAuth(username, pwd) {
        this.customConfig = Object.assign(Object.assign({}, this.customConfig), { basicAuth: {
                username: username,
                pass: pwd,
            } });
        return this;
    }
    // CATCH ALL CUSTOM CONFIG
    apply(namespace, prop) {
        const baseConfigKeys = Reflect.ownKeys(this.get());
        if (!baseConfigKeys.includes(namespace)) {
            this.customConfig = Object.assign(Object.assign({}, this.customConfig), { [namespace]: prop });
            return this;
        }
        else {
            throw new Error("BaseConfig key conflict");
        }
    }
    // adds offeringName and dashboardServiceApi to app config
    useDashboardNotice(offeringId) {
        this.customConfig = Object.assign(Object.assign({}, this.customConfig), { dashboardService: {
                offeringId,
                api: getStringEnvVar("DASHBOARD_SERVICE_API"),
            } });
        return this;
    }
    get() {
        return Object.freeze(Object.assign(Object.assign({}, this.baseConfig), this.customConfig));
    }
}
// Dynamic config
export const ApplicationConfig = new Config();
if (isTest || isRemoteTest) {
    ApplicationConfig.usePolicies(getStringEnvVar("POLICIES_API", ""), getStringEnvVar("BASICAUTH_USERNAME", ""), getStringEnvVar("BASICAUTH_PASS", ""), getStringEnvVar("SERVICE_ID", ""));
    ApplicationConfig.useServices({
        github: "https://api.github.com",
        hskV1: "https://ies-hsk-management.dev.identity-services.intranet.ibm.com/api/v1",
    });
}
export const baseConfig = ApplicationConfig.get();
