import { NextFunction, Request, Response } from "express";
/**
 * Router-level middleware used to ensure a request has been authenticated.
 *
 * If the request has not been authenticated and if the request contains an
 * authorization header, then it implies that this template is being used as
 * a pure backend service and the access token needs to be verified via
 * introspection.
 *
 * Next can be optionally passed as an argument, allowing it to be called outside
 * of this middleware file for manual authentication elsewhere.
 *
 * @type {void}
 * @param {Request} req
 * @param {Response} res
 * @param {NextFunction} next
 */
export declare const ensureAuthenticated: (req: Request, res: Response, next?: NextFunction | undefined) => Promise<void | boolean>;
/**
 * Router-level middleware used to ensure a request has been authenticated.
 *
 * If the request has not been authenticated, request will be redirected to the requests originalUrl
 *
 * @type {void}
 * @param {Request} req
 * @param {Response} res
 * @param {NextFunction} next
 */
export declare const ensureSessionAuthRedirectUrl: (req: Request, res: Response, next: NextFunction) => void;
/**
 * Router-level middleware used to ensure a request has been authenticated.
 *
 * If the request has not been authenticated, HTTP401Error will be thrown.
 *
 * @type {void}
 * @param {Request} req
 * @param {Response} res
 * @param {NextFunction} next
 */
export declare const ensureAuthUser: (req: Request, res: Response, next: NextFunction) => void;
/**
 * Router-level middleware used to validate if a requesting user has the configured admin blue group
 *
 * If the requesting user does not have the configured admin blue group, HTTP403Error will be thrown.
 *
 * @type {void}
 * @param {Request} req
 * @param {Response} res
 * @param {NextFunction} next
 */
export declare const ensureAdminUser: (req: Request, res: Response, next: NextFunction) => void;
/**
 * Router-level middleware used to validate a request against the basic auth credentials stored in the server config
 *
 * If credentials are not supplied, or are incorrect, HTTP403Error will be thrown. If a developer doesn't provide basicAuth
 * in the config, an error will get thrown. The desired behavior in this case is to have the server return 500.
 *
 * @type {void}
 * @param {Request} req
 * @param {Response} res
 * @param {NextFunction} next
 */
export declare const ensureBasicAuth: (req: Request, res: Response, next: NextFunction) => void;
/**
 * Router-level middleware used to ensure a user has accepted the applications global policy
 *
 * If the request does not have a user property, the request will be redirected to AUTH_ROUTES.LOGIN allow the user to login.
 *
 * If the user making the request has not accepted the applications policy, validateUserWithServiceIdApi, the request will be redirected to the req.requestUrl
 *
 * @type {void}
 * @param {Request} req
 * @param {Response} res
 * @param {NextFunction} next
 */
export declare const requirePolicy: (req: Request, res: Response, next: NextFunction) => Promise<void>;
