import fs from "fs";
import connectRedis from "connect-redis";
import session from "express-session";
import { baseConfig, isProduction } from "../config";
// import { RequestWithId } from "./requestId.middleware";
import { RedisClient } from "./client/RedisClient";
import { logMiddlewareStart } from "./middleware.api";
const createRedisSessionStore = () => {
    const RedisStore = connectRedis(session);
    const redisOptions = {
        host: baseConfig.redis.url
            ? new URL(baseConfig.redis.url).hostname
            : "redis",
        port: baseConfig.redis.port,
    };
    if (baseConfig.redis.cert) {
        redisOptions.tls = {
            // Refer to `tls.connect()` section in
            // https://nodejs.org/api/tls.html
            // for all supported options
            ca: fs.readFileSync(`cert/${baseConfig.redis.cert}`, "utf8"),
        };
    }
    const redisClient = new RedisClient(redisOptions);
    return new RedisStore({
        client: redisClient.client,
        ttl: 86400,
    });
};
export const handleSessions = (router) => {
    logMiddlewareStart("Sessions");
    /**
     * Option Descriptions
     *
     * - "secret"
     * This is a required option for the secret to sign the session ID cookie.
     * It can be a string or an array of multiple string secrets.
     *
     * - "resave"
     * Forces the session to be saved back to the session store even if the session was never modified during the request.
     * This may create race conditions when a client makes 2 parallel requests to our server
     * and changes made to the session on one request may be overwritten when the other request ends
     *
     * - "saveUninitalized"
     * Force a session that’s uninitialized to be saved to the store. A session is uninitialized when it’s new but not
     * modified. Setting this to false reduces server storage usage and comply with laws that require permission
     * before storing cookies.
     *
     * - "store"
     * The session store instance. It defaults to a new MemoryStore instance which is not production safe.
     *
     */
    const sessionConfig = {
        secret: baseConfig.session.secret,
        resave: false,
        cookie: {
            secure: false,
        },
        saveUninitialized: true,
        // TODO: Investiage why genId was breaking session state
        // genid: (req: Request): string => (req as RequestWithId).id as string,
    };
    // Customize as needed for your environment
    if (isProduction) {
        router.set("trust proxy", 1);
        if (sessionConfig.cookie) {
            sessionConfig.cookie.secure = true; // needed for HTTPS
        }
        // Uncomment to use Redis, Comment to use Express Session store
        sessionConfig.store = createRedisSessionStore();
    }
    router.use(session(sessionConfig));
};
