import fetch, { RequestInfo, RequestInit, Response } from "node-fetch";
import { Omit } from "utility-types";
import { ExternalApiError } from "../../common/rest/errors/externalApi.error";
export interface APIRequester {
    fetchJson: (url: RequestInfo, options?: RequestInit, errorResponseTransformer?: ApiErrorResponseTransformer) => Promise<any>;
    fetchJsonWithoutDefaultContentType: (url: RequestInfo, options?: RequestInit) => Promise<any>;
}
declare type RequestOptions = Omit<RequestInit, "method">;
declare type ApiErrorResponseTransformer = (res: Response) => Promise<ExternalApiError>;
/**
 *  Utility that wraps fetch to simplify its use with JSON-based APIs.
 *  TODO consider making a HttpApiJsonRequesterBuilder
 *  TODO consider making a RequestBuilder from the static methods
 */
export declare class HttpApiJsonRequester implements APIRequester {
    private static readonly LOGGER;
    private static readonly defaultOptions;
    private static readonly headers;
    private externalDefaultInstanceOptions;
    private apiErrorResponseTransformer?;
    private request;
    get fetch(): typeof fetch;
    set fetch(f: typeof fetch);
    private apiError;
    constructor(api: string, defaultOptions?: RequestInit, enableTracing?: boolean, apiErrorResponseTransformer?: ApiErrorResponseTransformer);
    /**
     * Apply ad-hoc scoped (hardcoded deep) merge for internal default props only
     * If default props become deeply nested consider using
     * a deep merge lib such as lodash _.merge
     * @param method
     * @param options
     */
    private static mergeProps;
    static getJson(options?: RequestOptions): Record<string, unknown>;
    static deleteJson(body?: {}, options?: RequestOptions): Record<string, unknown>;
    static postJson(body?: {}, options?: RequestOptions): Record<string, unknown>;
    static putJson(body?: {}, options?: RequestOptions): Record<string, unknown>;
    static patchJson(body?: {}, options?: RequestOptions): Record<string, unknown>;
    static oauth2Header(accessToken: string): {
        Authorization: string;
    };
    static basicAuth(username: string, password: string): {
        Authorization: string;
    };
    get error(): (status: number, message: Record<string, unknown> | undefined | string) => ExternalApiError;
    /**
     * Enhances fetch() as follows:
     * 1. It retrieves the body of a success response, parses it, and resolves to the resulting object.
     * 2. It rejects on receiving an error response, with an error that captures key reponse details.
     * 3. It applies default options if none are specified.
     * @param url the request url
     * @param options the fetch() configuration options
     * @param customApiErrorResponseTransformer optional Error Transformer that can be used in liue of throwing the default this.apiError
     */
    fetchJson(url: RequestInfo, options?: RequestInit, customApiErrorResponseTransformer?: ApiErrorResponseTransformer): Promise<string | Record<string, unknown>>;
    /**
     * Enhances fetchJson() as follows:
     * 1. It sends requests without the HttpApiJsonRequester.headers property
     * @param url the request url
     * @param options the fetch() configuration options
     * @param customApiErrorResponseTransformer optional Error Transformer that can be used in liue of throwing the default this.apiError
     */
    fetchJsonWithoutDefaultContentType(url: RequestInfo, options?: RequestInit, customApiErrorResponseTransformer?: ApiErrorResponseTransformer): Promise<string | Record<string, unknown>>;
}
export {};
