var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
import fetch from "node-fetch";
import { logger } from "../../common/logger";
import { ExternalApiError } from "../../common/rest/errors/externalApi.error";
import { traceFetchRequests } from "./apiTracer.utils";
/**
 *  Utility that wraps fetch to simplify its use with JSON-based APIs.
 *  TODO consider making a HttpApiJsonRequesterBuilder
 *  TODO consider making a RequestBuilder from the static methods
 */
export class HttpApiJsonRequester {
    constructor(api, defaultOptions = {}, enableTracing = true, apiErrorResponseTransformer) {
        this.externalDefaultInstanceOptions = defaultOptions;
        this.request = enableTracing ? traceFetchRequests(api) : fetch;
        this.apiErrorResponseTransformer = apiErrorResponseTransformer;
        this.apiError = ExternalApiError.factory(api);
    }
    get fetch() {
        return this.request;
    }
    set fetch(f) {
        this.request = f;
    }
    /**
     * Apply ad-hoc scoped (hardcoded deep) merge for internal default props only
     * If default props become deeply nested consider using
     * a deep merge lib such as lodash _.merge
     * @param method
     * @param options
     */
    static mergeProps(method, options = {}, useDefaultContentType = true) {
        return Object.assign(Object.assign(Object.assign({}, HttpApiJsonRequester.defaultOptions), options), { method, headers: Object.assign(Object.assign(Object.assign({}, HttpApiJsonRequester.defaultOptions.headers), (useDefaultContentType ? HttpApiJsonRequester.headers : {})), options.headers) });
    }
    static getJson(options = {}) {
        return this.mergeProps("GET", options);
    }
    static deleteJson(body = {}, options = {}) {
        return this.mergeProps("DELETE", Object.assign(Object.assign({}, options), { body: JSON.stringify(body) }));
    }
    static postJson(body = {}, options = {}) {
        return this.mergeProps("POST", Object.assign(Object.assign({}, options), { body: JSON.stringify(body) }));
    }
    static putJson(body = {}, options = {}) {
        return this.mergeProps("PUT", Object.assign(Object.assign({}, options), { body: JSON.stringify(body) }));
    }
    static patchJson(body = {}, options = {}) {
        return this.mergeProps("PATCH", Object.assign(Object.assign({}, options), { body: JSON.stringify(body) }));
    }
    static oauth2Header(accessToken) {
        return {
            Authorization: `Bearer ${accessToken}`,
        };
    }
    static basicAuth(username, password) {
        return {
            Authorization: "Basic " + Buffer.from(`${username}:${password}`).toString("base64"),
        };
    }
    get error() {
        return this.apiError;
    }
    /**
     * Enhances fetch() as follows:
     * 1. It retrieves the body of a success response, parses it, and resolves to the resulting object.
     * 2. It rejects on receiving an error response, with an error that captures key reponse details.
     * 3. It applies default options if none are specified.
     * @param url the request url
     * @param options the fetch() configuration options
     * @param customApiErrorResponseTransformer optional Error Transformer that can be used in liue of throwing the default this.apiError
     */
    fetchJson(url, options = {}, customApiErrorResponseTransformer) {
        const instanceOptions = Object.assign(Object.assign(Object.assign({}, this.externalDefaultInstanceOptions), options), { headers: Object.assign(Object.assign({}, this.externalDefaultInstanceOptions.headers), options.headers) });
        const allOptions = HttpApiJsonRequester.mergeProps(instanceOptions.method, instanceOptions);
        const toResponseMessage = (response) => __awaiter(this, void 0, void 0, function* () {
            var _a;
            const contentType = (_a = response.headers) === null || _a === void 0 ? void 0 : _a.get("content-type");
            const status = response.status;
            if (!contentType || status === 204) {
                return;
            }
            if (contentType.indexOf("application/json") !== -1) {
                return response.json();
            }
            else {
                return response.text();
            }
        });
        return this.fetch(url, allOptions).then((response) => __awaiter(this, void 0, void 0, function* () {
            const errorTransformer = customApiErrorResponseTransformer
                ? customApiErrorResponseTransformer
                : this.apiErrorResponseTransformer;
            if (!response.ok) {
                HttpApiJsonRequester.LOGGER.info(yield toResponseMessage(response));
                if (errorTransformer) {
                    HttpApiJsonRequester.LOGGER.debug("Using the custom http error response transformer");
                    throw yield errorTransformer(response);
                }
                throw this.apiError(response.status, response.statusText);
            }
            return toResponseMessage(response);
        }));
    }
    /**
     * Enhances fetchJson() as follows:
     * 1. It sends requests without the HttpApiJsonRequester.headers property
     * @param url the request url
     * @param options the fetch() configuration options
     * @param customApiErrorResponseTransformer optional Error Transformer that can be used in liue of throwing the default this.apiError
     */
    fetchJsonWithoutDefaultContentType(url, options = {}, customApiErrorResponseTransformer) {
        const instanceOptions = Object.assign(Object.assign(Object.assign({}, this.externalDefaultInstanceOptions), options), { headers: Object.assign(Object.assign({}, this.externalDefaultInstanceOptions.headers), options.headers) });
        const allOptions = HttpApiJsonRequester.mergeProps(instanceOptions.method, instanceOptions, false);
        const toResponseMessage = (response) => __awaiter(this, void 0, void 0, function* () {
            var _a;
            const contentType = (_a = response.headers) === null || _a === void 0 ? void 0 : _a.get("content-type");
            if (!contentType) {
                return;
            }
            if (contentType.indexOf("application/json") !== -1) {
                return response.json();
            }
            else {
                return response.text();
            }
        });
        return this.fetch(url, allOptions).then((response) => __awaiter(this, void 0, void 0, function* () {
            const errorTransformer = customApiErrorResponseTransformer
                ? customApiErrorResponseTransformer
                : this.apiErrorResponseTransformer;
            if (!response.ok) {
                HttpApiJsonRequester.LOGGER.info(yield toResponseMessage(response));
                if (errorTransformer) {
                    HttpApiJsonRequester.LOGGER.debug("Using the custom http error response transformer");
                    throw yield errorTransformer(response);
                }
                throw this.apiError(response.status, response.statusText);
            }
            return toResponseMessage(response);
        }));
    }
}
HttpApiJsonRequester.LOGGER = logger.child("HttpApiJsonRequester");
HttpApiJsonRequester.defaultOptions = { compress: false };
HttpApiJsonRequester.headers = { "Content-Type": "application/json" };
