import { ValidationError } from "class-validator";
import { NextFunction, Request, Response } from "express";
import { validationResult } from "express-validator";
import { NO_CONTENT, OK } from "http-status-codes";

import { BusinessViolationCode } from "../businessRules/businessViolationCodes";
import { Model } from "../data/model.api";
import { HTTP405Error } from "./errors/methodNotAllowed.error";
import { HTTP422Error } from "./errors/validation/businessViolation.error";
import { RestResponse } from "./restResponse";

type MiddlewareMethod = (
  req: Request,
  res: Response,
  next: NextFunction,
  param: string
) => any;

type IsControllerMiddlewareMethod<T> = T extends MiddlewareMethod
  ? T
  : MiddlewareMethod;

export type ControllerMiddlewareMethod = IsControllerMiddlewareMethod<
  keyof AbstractRestController<any>
>;

export abstract class AbstractRestController<T extends Model> {
  respond(
    res: Response,
    item: T | T[] | undefined,
    statusCode: number = OK
  ): Response {
    const response = new RestResponse(item);
    return res.status(statusCode).json(response);
  }

  respondNoContent(res: Response, statusCode: number = NO_CONTENT): Response {
    return res.status(statusCode).json();
  }

  throwMethodNotAllowedError(req: Request, res: Response) {
    throw new HTTP405Error();
  }

  throwBusinessViolation(
    businessViolationCode: BusinessViolationCode,
    message?: string,
    errors?: Record<string, unknown>[]
  ) {
    throw new HTTP422Error(businessViolationCode, message, errors);
  }

  /**
   * Validates the model using class-validator library
   * @param model
   */
  async validateModel(model: Model): Promise<void> {
    if (!model.isValid()) {
      const invalidations = await model.validate();

      const message = invalidations.reduce(
        (msg: string, validationError: ValidationError) => {
          return `${msg}\n${validationError.toString(true)}`;
        },
        ""
      );

      this.throwBusinessViolation(
        BusinessViolationCode.DATA_CONSTRAINT,
        message
      );
    }
  }

  /**
   * Validates the request against any express-validator library middleware
   * that may have been applied on the route that the request came through
   * @param req
   */
  validateRequest(req: Request): void {
    const result = validationResult(req);

    if (!result.isEmpty()) {
      this.throwBusinessViolation(
        BusinessViolationCode.DATA_CONSTRAINT,
        "Request format is invalid",
        result.array()
      );
    }
  }
}
