import { getBooleanEnvVar, getIntegerEnvVar, getStringEnvVar } from "../utils";
import { ApplicationConfig } from "./env.config";

describe("ENV Config", () => {
  describe("ApplicationConfig", () => {
    describe("#get", () => {
      it("should return an object with the correct defaults set", () => {
        const defaultConfig = {
          app: {
            serviceId: getStringEnvVar("APP_NAME", "ies-node-api-template"),
            serviceType: getStringEnvVar("SERVICE_TYPE", "BFF"),
            apiBasePath: getStringEnvVar("API_BASE_PATH", "/api"),
            environment: "test",
            port: getIntegerEnvVar("PORT", 8443),
            serveStatic: getBooleanEnvVar("SERVE_STATIC", false),
            swagger: getBooleanEnvVar("SWAGGER", false),
            logLevel: getStringEnvVar("LOG_LEVEL", "info"),
            unauthorizedRedirectUrl: getStringEnvVar(
              "UNAUTHORIZED_REDIRECT_URL",
              "https://ies-portal.prod.identity-services.intranet.ibm.com/#/"
            ),
            enableHttpRequestLogging: getBooleanEnvVar(
              "ENABLE_HTTP_REQUEST_LOGGING",
              true
            ),
          },
          health: {
            checkUrl: getStringEnvVar("HEALTH_CHECK_URL", "/actuator/health"),
            beforeShutdownTimeout: getIntegerEnvVar(
              "TIMEOUT_BEFORE_SHUTDOWN",
              6000
            ),
          },
          session: {
            secret: getStringEnvVar("SESSION_SECRET", "nodeapi"),
          },
          redis: {
            // EFFECTS TEMPLATE session middleware
            url: getStringEnvVar("REDIS_URL", ""),
            port: getIntegerEnvVar("REDIS_PORT", 6379),
            cert: getStringEnvVar("REDIS_CERT_FILE", "redis.cer"),
            prefix: getStringEnvVar("REDIS_PREFIX", "ies-node-api-template"),
          },
          mongodb: {
            url: getStringEnvVar("MONGODB_URL", ""),
          },
          openid: {
            client_id: getStringEnvVar("OPENID_CLIENT_ID", "_"), // get from SSO Provisioner
            client_secret: getStringEnvVar("OPENID_CLIENT_SECRET", "_"), // get from SSO Provisioner
            discovery_url: getStringEnvVar(
              "OPENID_DISCOVERY_URL",
              "https://preprod.login.w3.ibm.com/oidc/endpoint/default/.well-known/openid-configuration"
            ), // default URL is for pre-prod environment
            callback_url: getStringEnvVar(
              "OPENID_CALLBACK_URL",
              "https://localhost:8443/api/openid_connect_login"
            ), // default URL is for local devlopment
            logout_url: getStringEnvVar(
              "OPENID_LOGOUT_URL",
              "https://w3id.alpha.sso.ibm.com/pkmslogout"
            ), // default URL is for pre-prod environment
            introspection_url: getStringEnvVar(
              "OPENID_INTROSPECTION_URL",
              "https://preprod.login.w3.ibm.com/oidc/endpoint/default/introspect"
            ), // default URL is for pre-prod environment
            token_url: getStringEnvVar(
              "OPENID_TOKEN_URL",
              "https://preprod.login.w3.ibm.com/v1.0/endpoint/default/token"
            ), // default URL to retrieve token is for pre-prod environment
          },
          bluegroups: {
            //default blue group
            admin: [getStringEnvVar("BLUEGROUPS_ADMIN", "")],
            auditor: [getStringEnvVar("BLUEGROUPS_AUDITOR", "")],
            appUser: [getStringEnvVar("BLUEGROUPS_APP_USER", "")],
            bypassAppUserAuth: getBooleanEnvVar("BYPASS_APP_USER_AUTH", false),
            devOps: [getStringEnvVar("BLUEGROUPS_DEVOPS", "")],
            po: [getStringEnvVar("BLUEGROUPS_PRODUCTOWNER", "")],
            external: [getStringEnvVar("BLUEGROUPS_EXTERNAL_USER", "")],
            teamMember: [getStringEnvVar("BLUEGROUPS_TEAM_MEMBER", "")],
          },
        };
        expect(ApplicationConfig.get()).toMatchObject(defaultConfig);
      });
    });
    describe("#useServices", () => {
      it("should add a new service to the api property of the base config object", () => {
        let config = ApplicationConfig.get();
        const apiKeyToTest = "apiKey";
        const newService = { [apiKeyToTest]: "https://key" };

        expect(Reflect.ownKeys(config.api).includes(apiKeyToTest)).toBeFalsy();

        ApplicationConfig.useServices(newService);
        config = ApplicationConfig.get();

        expect(Reflect.ownKeys(config.api).includes(apiKeyToTest)).toBeTruthy();
        expect(config.api[apiKeyToTest]).toBe(newService[apiKeyToTest]);
      });
    });
    describe("#addBluegroups", () => {
      it("should add a new bluegroupp to the bluegroup property of the base config object", () => {
        let config = ApplicationConfig.get();
        const bluegroupdToTest = "tester";
        const newBlueGroup = { [bluegroupdToTest]: "BLUEGROUPS_TESTER" };

        expect(
          Reflect.ownKeys(config.bluegroups).includes(bluegroupdToTest)
        ).toBeFalsy();

        ApplicationConfig.addBluegroups(newBlueGroup);
        config = ApplicationConfig.get();

        expect(
          Object.keys(config.bluegroups).includes(bluegroupdToTest)
        ).toBeTruthy();
        expect(config.bluegroups[bluegroupdToTest]).toBe(
          newBlueGroup[bluegroupdToTest]
        );
      });
      it("should add new blue group role that is a string to an existing default bluegroup", () => {
        const newBlueGroup = { admin: "BLUEGROUPS_ADMIN" };
        const initialAdminGroups = ApplicationConfig.get().bluegroups.admin;
        const expectAdminGroupsBeforeAdd = ["ADMIN"];
        const expectAdminGroupsAfterAdd = ["ADMIN", "BLUEGROUPS_ADMIN"];

        expect(initialAdminGroups).toMatchObject(expectAdminGroupsBeforeAdd);

        ApplicationConfig.addBluegroups(newBlueGroup);
        const postAdminGroups = ApplicationConfig.get().bluegroups.admin;

        expect(postAdminGroups).toMatchObject(expectAdminGroupsAfterAdd);
      });

      it("should add new blue group role that is an array to an existing default bluegroup", () => {
        const newBlueGroup = { auditor: ["AN_AUDITOR", "ANOTHER_GROUP"] };
        const initialAdminGroups = ApplicationConfig.get().bluegroups.auditor;
        const expectAdminGroupsBeforeAdd = ["AUDITOR"];
        const expectAdminGroupsAfterAdd = [
          "AUDITOR",
          "AN_AUDITOR",
          "ANOTHER_GROUP",
        ];

        expect(initialAdminGroups).toMatchObject(expectAdminGroupsBeforeAdd);

        ApplicationConfig.addBluegroups(newBlueGroup);
        const postAdminGroups = ApplicationConfig.get().bluegroups.auditor;

        expect(postAdminGroups).toMatchObject(expectAdminGroupsAfterAdd);
      });
    });
    describe("#apply", () => {
      it("should add a new property base config object", () => {
        let config = ApplicationConfig.get();
        const newKey = "test";
        const newProp = "thisIsATest";

        expect(Reflect.ownKeys(config).includes(newKey)).toBeFalsy();

        ApplicationConfig.apply(newKey, newProp);
        config = ApplicationConfig.get();

        expect(Object.keys(config).includes(newKey)).toBeTruthy();
        expect(config[newKey]).toBe(newProp);
      });
      it("should throw an error when a developer attempts to overwrite one of the default baseConfig keys", () => {
        const newProp = {
          openid: {
            client_id: "test1234",
          },
        };
        const testProp = "openid";
        const errorTest = jest.fn(() => {
          ApplicationConfig.apply(testProp, newProp);
        });
        expect(errorTest).toThrowError();
      });
    });
    describe("common config", () => {
      it("should allow a developer to add both enableMaintenanceMode and usePolicies", () => {
        let config = ApplicationConfig.get();
        const policiesApi = "path";
        const servicesApi = "path-two";
        const serviceId = "an-app";
        const testBasicAuth = {
          user: "demarcuskennedy95@gmail.com",
          pass: "abc123",
        };
        const testPoliciesObj = {
          policiesApi,
          serviceId,
        };

        const testMaintenanceRecord = {
          servicesApi,
          serviceId,
        };

        ApplicationConfig.usePolicies(
          testPoliciesObj.policiesApi,
          testBasicAuth.user,
          testBasicAuth.pass,
          testPoliciesObj.serviceId
        );
        ApplicationConfig.enableMaintenanceMode(
          testMaintenanceRecord.servicesApi,
          testBasicAuth.user,
          testBasicAuth.pass,
          testMaintenanceRecord.serviceId
        );
        config = ApplicationConfig.get();

        expect(Object.keys(config).includes("common")).toBeTruthy();

        const expected = {
          policiesApi,
          servicesApi,
          serviceId,
        };
        expect(config.common).toMatchObject(expected);
      });
      describe("#usePolicies", () => {
        it("should allow a developer to add policy configuration", () => {
          let config = ApplicationConfig.get();
          const testBasicAuth = {
            user: "demarcuskennedy95@gmail.com",
            pass: "abc123",
          };
          const testPoliciesObj = {
            policiesApi: "path",
            serviceId: "an-app",
          };

          ApplicationConfig.usePolicies(
            testPoliciesObj.policiesApi,
            testBasicAuth.user,
            testBasicAuth.pass,
            testPoliciesObj.serviceId
          );
          config = ApplicationConfig.get();

          expect(Object.keys(config).includes("common")).toBeTruthy();
          expect(config.common).toMatchObject(testPoliciesObj);
        });
      });
      describe("#enableMaintenanceMode", () => {
        it("should allow a developer to add maintenanceMode configuration", () => {
          let config = ApplicationConfig.get();
          const testBasicAuth = {
            user: "demarcuskennedy95@gmail.com",
            pass: "abc123",
          };
          const testMaintenanceRecord = {
            servicesApi: "path",
            serviceId: "an-app",
          };

          ApplicationConfig.enableMaintenanceMode(
            testMaintenanceRecord.servicesApi,
            testBasicAuth.user,
            testBasicAuth.pass,
            testMaintenanceRecord.serviceId
          );
          config = ApplicationConfig.get();

          expect(Object.keys(config).includes("common")).toBeTruthy();
          expect(config.common).toMatchObject(testMaintenanceRecord);
        });
      });
    });
    describe("#setBasicAuth", () => {
      it("should add a new basicAuth property to the base config object ", () => {
        let config = ApplicationConfig.get();
        const testBasicAuth = {
          username: "demarcuskennedy95@gmail.com",
          pass: "abc123",
        };

        ApplicationConfig.setBasicAuth(
          testBasicAuth.username,
          testBasicAuth.pass
        );
        config = ApplicationConfig.get();

        expect(Object.keys(config).includes("basicAuth")).toBeTruthy();
        expect(config.basicAuth).toMatchObject(testBasicAuth);
      });
    });
    describe("#useDashboardNotice", () => {
      let config = ApplicationConfig.get();
      const testOfferingId = "6102eaa24908dd09cddfee39";
      const testOfferingObj = {
        offeringId: testOfferingId,
        api: "https://ies-dashboard-service.dev.identity-services.intranet.ibm.com/api/v1",
      };

      ApplicationConfig.useDashboardNotice(testOfferingId);
      config = ApplicationConfig.get();

      expect(Object.keys(config).includes("dashboardService")).toBeTruthy();
      expect(config.dashboardService).toMatchObject(testOfferingObj);
    });
  });
});
