import https from "https";

import { IResolvers } from "@graphql-tools/utils";
import { Request, Response } from "express";
import { DocumentNode } from "graphql";

import {
  getBooleanEnvVar,
  getIntegerEnvVar,
  getStringEnvVar,
} from "../utils/env.utils";

/**
 * Set default NODE ENV variables
 */

// Allows for certs to work with HTTPS locally
process.env.NODE_TLS_REJECT_UNAUTHORIZED = "0";

/**
 * Process environment
 */

type apiOptions = {
  [key: string]: string;
};

type blueGroupOptions = {
  [key: string]: string | string[];
};

const environment = getStringEnvVar("NODE_ENV", "development");

export const isProduction = environment === "production";
export const isDevelopment = environment === "development";
export const isTest = environment === "test"; // Defines that we are running tests, but locally
export const isRemoteTest = environment === "remote-test"; // Defines that we are running tests, but externally (like in a deployment pipeline)

// Set test specific environments for testsuite
// TODO find away to include this elsewhere
if (isTest || isRemoteTest) {
  process.env.BLUEGROUPS_ADMIN = "ADMIN";
  process.env.BLUEGROUPS_AUDITOR = "AUDITOR";
  process.env.BLUEGROUPS_APP_USER = "NODE_TEMPLATE";
  process.env.BLUEGROUPS_DEVOPS = "DEVOPS";
  process.env.BLUEGROUPS_PRODUCTOWNER = "PRODUCT_OWNER";
  process.env.BLUEGROUPS_EXTERNAL_USER = "EXTERNAL";
  process.env.BLUEGROUPS_TEAM_MEMBER = "TEAM_MEMBER";
  process.env.LOG_LEVEL = "silent";
  process.env.DASHBOARD_SERVICE_API =
    "https://ies-dashboard-service.dev.identity-services.intranet.ibm.com/api/v1";
  process.env.OFFERING_ID = "6102eaa24908dd09cddfee39";
}

if (isRemoteTest) {
  process.env.HSK_API =
    "https://ies-hsk-management.dev.identity-services.intranet.ibm.com/api/v1";
  process.env.NODE_TLS_REJECT_UNAUTHORIZED = "0";
  process.env.OPENID_CALLBACK_URL = "default_callback_url";
  process.env.OPENID_CLIENT_ID = "default_client_id";
  process.env.OPENID_CLIENT_SECRET = "default_client_secret";
  process.env.OPENID_DISCOVERY_URL = "default_discovery_url";
  process.env.OPENID_LOGOUT_URL = "default_logout_url";
  process.env.POLICIES_API = "default_policies_api";
  process.env.BASICAUTH_PASS = "default_BASICAUTH_PASS";
  process.env.SERVICE_ID = "ies-hsk-registration";
  process.env.BASICAUTH_USERNAME = "default_BASICAUTH_USERNAME";
  process.env.PORT = "8443";
  process.env.REDIS_CERT_FILE = "redis.cer";
  process.env.REDIS_PORT = "31440";
  process.env.REDIS_PREFIX = "ies-hsk-registration-frontend:";
  process.env.REDIS_URL = "redis_url";
  process.env.SERVE_STATIC = "true";
  process.env.UNAUTHORIZED_REDIRECT_URL = "default_unauthorized_redirect_url";
  process.env.W3_SUPPORT_PAGE = "https://w3.ibm.com/help/#/article/hsk";
}

class Config {
  private baseConfig: any = Object.freeze({
    app: {
      serviceId: getStringEnvVar("APP_NAME", "ies-node-api-template"), // Name for this service, not to be confused with serviceId optionally under common - could potentially merge these two together in the future
      serviceType: getStringEnvVar("SERVICE_TYPE", "BFF"), // Modifies notAuthenticatedResponse behavior, will redirect users without an auth header to login route when set to BFF, otherwise throws 401 HTTP Error
      apiBasePath: getStringEnvVar("API_BASE_PATH", "/api"), // Defines the base path for all API (non-middleware) routes - The default sets the routes defined in services to /api/{your-service}
      environment, // What environment the server is running in, useful when running tests
      port: getIntegerEnvVar("PORT", 8443), // Port that the server is listening on
      serveStatic: getBooleanEnvVar("SERVE_STATIC", false), // Boolean that tells the server whether or not to serve-static content (see serveStatic middleware)
      enableMaintenanceMode: getBooleanEnvVar("ENABLE_MAINTENANCE_MODE", false), // modifies the serve-static method to check for maintenance mode first, requires additional information via enableMaintenanceMode and serveStatic must be true
      swagger: getBooleanEnvVar("SWAGGER", false), // Boolean that indicates whether or not this application has swagger documentation/ui. Requires additional information when instantiating the server to work. See readme
      logLevel: getStringEnvVar("LOG_LEVEL", "info"), // Modifies pino-logger behavior, will only show logs at the given sev level and above
      unauthorizedRedirectUrl: getStringEnvVar(
        "UNAUTHORIZED_REDIRECT_URL",
        "https://ies-portal.prod.identity-services.intranet.ibm.com/#/"
      ), // Exposes an unauthorized redirect route for OIDC login flow
      enableHttpRequestLogging: getBooleanEnvVar(
        "ENABLE_HTTP_REQUEST_LOGGING",
        true
      ), // Boolean that controls whether or not to enable logging
      useHttps: getBooleanEnvVar("USE_HTTPS", true), // Boolean that controls whether the server runs in http or https mode
      requireLogin: getBooleanEnvVar("REQUIRE_LOGIN", true), // Boolean that determines whether or not handleLogin middleware is automatically included in auth middleware (OIDC flow)
    },
    health: {
      checkUrl: getStringEnvVar("HEALTH_CHECK_URL", "/actuator/health"), // Determines the api endpoint that the health actuator exists on
      beforeShutdownTimeout: getIntegerEnvVar("TIMEOUT_BEFORE_SHUTDOWN", 6000), // modifies terminus behavior, see health.ts file for more information
    },
    session: {
      secret: getStringEnvVar("SESSION_SECRET", "nodeapi"), // This is a required option for the secret to sign the session ID cookie, it can be a string or an array of multiple string secrets.. see session.middleware
    },
    redis: {
      // EFFECTS TEMPLATE session middleware - see session.middleware.ts
      url: getStringEnvVar("REDIS_URL", ""), // url for the redis session connection
      port: getIntegerEnvVar("REDIS_PORT", 6379), // port for the redis session
      cert: getStringEnvVar("REDIS_CERT_FILE", "redis.cer"), // used to construct the filepath which reads the tls cert
      prefix: getStringEnvVar("REDIS_PREFIX", "ies-node-api-template"), // sets the keyPrefix option for redis, see RedisClient
    },
    mongodb: {
      url: getStringEnvVar("MONGODB_URL", ""), // url for mongodb, if used
    },
    // Config variables related to the openid (oidc) login flow - strategy is defined in OpenIdConnect strategy as scope "openid" response_type "code"
    // Additional certificates may be required to make this login flow work, see certs defined in the strategy
    // In order to make oidc flow work, the variables here that are consumed by the strategy must be present and correct. The other oidc variables are used for other purposes
    // Personally, I found this helpful in understanding our login flow - https://darutk.medium.com/diagrams-of-all-the-openid-connect-flows-6968e3990660#:~:text=1.%20response_type%3Dcode
    openid: {
      client_id: getStringEnvVar("OPENID_CLIENT_ID", "_"), // the consuming applications client_id, get from SSO Provisioner - defaulting to '_' for OpenIDConnectStrategy typing
      client_secret: getStringEnvVar("OPENID_CLIENT_SECRET", "_"), // the consuming applications client_secret, get from SSO Provisioner - defaulting to '_' for OpenIDConnectStrategy typing
      discovery_url: getStringEnvVar(
        "OPENID_DISCOVERY_URL",
        "https://preprod.login.w3.ibm.com/oidc/endpoint/default/.well-known/openid-configuration"
      ), // The discovery_url for the strategy, I believe that this refers to the authorization server
      callback_url: getStringEnvVar(
        "OPENID_CALLBACK_URL",
        "https://localhost:8443/api/openid_connect_login"
      ), // callback url for the strategy, usually is https://{your-app-here}/api/openid_connect_login
      logout_url: getStringEnvVar(
        "OPENID_LOGOUT_URL",
        "https://w3id.alpha.sso.ibm.com/pkmslogout"
      ), // logout url to destroy the user session
      introspection_url: getStringEnvVar(
        "OPENID_INTROSPECTION_URL",
        "https://preprod.login.w3.ibm.com/oidc/endpoint/default/introspect"
      ), // not consumed by the strategy, this endpoint is involved in verifying a users token (if one exists) against the introspection (auth?) server
      token_url: getStringEnvVar(
        "OPENID_TOKEN_URL",
        "https://preprod.login.w3.ibm.com/v1.0/endpoint/default/token"
      ),
      issuer_url: getStringEnvVar(
        "OPENID_ISSUER_ID",
        "https://preprod.login.w3.ibm.com/oidc/endpoint/default"
      ),
      authorization_url: getStringEnvVar(
        "OPENID_AUTHORIZATION_URL",
        "https://preprod.login.w3.ibm.com/oidc/endpoint/default/authorize"
      ),
      user_info_url: getStringEnvVar(
        "OPENID_USER_INFO_URL",
        "https://preprod.login.w3.ibm.com/v1.0/endpoint/default/userinfo"
      ),

      // not consumed by the strategy, as our strategy (I believe) abstracts this away for us via use of the discovery_url (or callback_url?), but if implementing this flow without using the strategy
      // (as is the case with Swagger), we need to define the endpoint to send our code to (once retrieved from the authorization server) so that we can receive a token back
    },
    bluegroups: {
      // bluegroups is defined on the return data from w3id, we use it to construct properties for our user model
      // e.g. if a user has reg_admin as a value in their bluegroups array and admin bluegroups is set to reg_admin
      // that user would have the isAdmin boolean set to true on their user model
      admin: [getStringEnvVar("BLUEGROUPS_ADMIN", "")], // defines admin permissions for the application
      auditor: [getStringEnvVar("BLUEGROUPS_AUDITOR", "")], // defines auditor permissions for the application
      appUser: [getStringEnvVar("BLUEGROUPS_APP_USER", "")], // defines appUser for the application (our team uses this to see if a user can even use our application, not always used)
      bypassAppUserAuth: getBooleanEnvVar("BYPASS_APP_USER_AUTH", false), // Boolean that bypasses restricting the application to users as defined by appUser in bluegroups, if set to true a user will need to have the appUser role to access the application (if automatic oidc login flow is enabled)
      devOps: [getStringEnvVar("BLUEGROUPS_DEVOPS", "")], // defines devops permissions for the application
      po: [getStringEnvVar("BLUEGROUPS_PRODUCTOWNER", "")], // defines product owner permissions for the application
      external: [getStringEnvVar("BLUEGROUPS_EXTERNAL_USER", "")], // defines external user role for the application (don't know if this is ever actually used)
      teamMember: [getStringEnvVar("BLUEGROUPS_TEAM_MEMBER", "")], // defines teamMember permissions for the application (if they belong to our team)
      helpDesk: [getStringEnvVar("BLUEGROUPS_HELP_DESK", "")], // defines IT Helpdesk permissions for the application
    },
  });

  // Anything not listed under the baseConfig gets added to the customConfig optionally by the user
  // This customConfig is needed to make certain features work with the template
  private customConfig: any = {};

  // adds 'api' prop to 'customConfig' object and spreads the values of the provided 'apis' object,
  useServices(apis: apiOptions) {
    this.customConfig = {
      ...this.customConfig,
      api: { ...apis },
    };
    return this;
  }

  // adds values to the 'bluegroups' prop of the 'customConfig' object - spreads the values of
  // the provided 'roles' object
  addBluegroups(roles: blueGroupOptions) {
    const baseConfigBlueGroups = Reflect.ownKeys(this.get().bluegroups);
    const newBluegroups = Reflect.ownKeys(roles).reduce(
      (acc, newBlueGroupKey: any) => {
        if (!baseConfigBlueGroups.includes(newBlueGroupKey)) {
          return {
            ...acc,
            [newBlueGroupKey]: roles[newBlueGroupKey],
          };
        } else {
          const newGroups = Array.isArray(roles[newBlueGroupKey])
            ? roles[newBlueGroupKey]
            : [roles[newBlueGroupKey]];

          return {
            ...acc,
            [newBlueGroupKey]: [
              ...this.get().bluegroups[newBlueGroupKey],
              ...newGroups,
            ],
          };
        }
      },
      {}
    );

    this.customConfig = {
      ...this.customConfig,
      bluegroups: { ...this.baseConfig.bluegroups, ...newBluegroups },
    };

    return this;
  }
  // EFFECTS TEMPLATE MIDDLEWARE
  usePolicies(api: string, username: string, pwd: string, serviceId: string) {
    this.customConfig = {
      ...this.customConfig,
      common: {
        ...this.customConfig?.common,
        policiesApi: api,
        serviceId: serviceId,
      },
    };

    this.setBasicAuth(username, pwd);

    return this;
  }

  // EFFECTS TEMPLATE MIDDLEWARE
  enableMaintenanceMode(
    api: string,
    username: string,
    password: string,
    serviceId: string
  ) {
    this.customConfig = {
      ...this.customConfig,
      common: {
        ...this.customConfig?.common,
        servicesApi: api,
        serviceId,
      },
    };

    this.setBasicAuth(username, password);

    return this;
  }

  // EFFECTS TEMPLATE MIDDLEWARE
  setBasicAuth(username: string, pwd: string) {
    this.customConfig = {
      ...this.customConfig,
      basicAuth: {
        username: username,
        pass: pwd,
      },
    };
    return this;
  }

  // CATCH ALL CUSTOM CONFIG
  apply(namespace: string, prop: any) {
    const baseConfigKeys = Reflect.ownKeys(this.get());
    if (!baseConfigKeys.includes(namespace)) {
      this.customConfig = { ...this.customConfig, [namespace]: prop };
      return this;
    } else {
      throw new Error("BaseConfig key conflict");
    }
  }

  // adds offeringName and dashboardServiceApi to app config
  useDashboardNotice(offeringId: string) {
    this.customConfig = {
      ...this.customConfig,
      dashboardService: {
        offeringId,
        api: getStringEnvVar("DASHBOARD_SERVICE_API"),
      },
    };

    return this;
  }

  get(): any {
    return Object.freeze({ ...this.baseConfig, ...this.customConfig });
  }
}

// Dynamic config
export const ApplicationConfig = new Config();

if (isTest || isRemoteTest) {
  ApplicationConfig.usePolicies(
    getStringEnvVar("POLICIES_API", ""),
    getStringEnvVar("BASICAUTH_USERNAME", ""),
    getStringEnvVar("BASICAUTH_PASS", ""),
    getStringEnvVar("SERVICE_ID", "")
  );
  ApplicationConfig.useServices({
    github: "https://api.github.com",
    hskV1:
      "https://ies-hsk-management.dev.identity-services.intranet.ibm.com/api/v1",
  });
}
export const baseConfig = ApplicationConfig.get();
// Static config
// Internal only for template; Don't export for consuming apps!
/**
 * Interface exposes only the parts of the config
 * above that is needed to dynamically configure
 * the core app logic
 */
export interface AppConfig {
  apiBasePath: string;
  port?: number;
  serveStatic?: boolean;
  enableMaintenanceMode?: boolean;
  swagger?: boolean;
  enableHttpRequestLogging?: boolean;
  logLevel?: string;
}

export interface AppServerConfig {
  port: number;
  https?: https.ServerOptions;
}

export type ApolloContextFn = (req: Request, res: Response) => any;

export interface GraphQLConfig {
  rootPath: string;
  typeDefs: DocumentNode[];
  resolvers: IResolvers[];
  context?: ApolloContextFn;
  uploads?: boolean;
}
