import { Router } from "express";
import swaggerJsdoc from "swagger-jsdoc";
import swaggerUi from "swagger-ui-express";

import { baseConfig, swaggerBasePath, SwaggerOptions } from "../config";
import { logMiddlewareStart } from "./middleware.api";

// SecuritySchemes are defined by official swagger documentation, see https://swagger.io/docs/specification/authentication/ for more information
// Not all definitions are included, feel free to add more here as necessary
interface ImplicitFlow {
  implicit: {
    authorizationUrl: string;
    scopes: Record<string, unknown>;
  };
}

interface AuthorizationCodeFlow {
  authorizationCode: {
    authorizationUrl: string;
    tokenUrl: string;
    scopes: Record<string, unknown>;
  };
}

interface PasswordFlow {
  password: {
    tokenUrl: string;
    scopes: Record<string, unknown>;
  };
}

interface ClientCredentialsFlow {
  clientCredentials: {
    tokenUrl: string;
    scopes: Record<string, unknown>;
  };
}

type flow =
  | AuthorizationCodeFlow
  | PasswordFlow
  | ClientCredentialsFlow
  | ImplicitFlow;

interface OAuth {
  [key: string]: {
    type: string;
    description: string;
    flows: flow;
  };
}

interface BearerAuth {
  [key: string]: {
    type: string;
    scheme: string;
    bearerFormat?: string;
  };
}

interface ApiKeyAuth {
  [key: string]: {
    type: string;
    in: "header" | "query" | "cookie";
    name: string;
  };
}

interface BasicAuth {
  [key: string]: {
    type: string;
    scheme: string;
  };
}

export type SecuritySchemes = OAuth | BearerAuth | ApiKeyAuth | BasicAuth;

export const handleAPIDocs = (
  router: Router,
  swaggerOptions = SwaggerOptions,
  uiOptions = { explorer: true } as Record<string, unknown>
): void => {
  logMiddlewareStart("Api Docs");

  const path = `${baseConfig.app.apiBasePath}${swaggerBasePath}`;
  const docs = swaggerJsdoc(swaggerOptions);

  router.use(path, swaggerUi.serve, swaggerUi.setup(docs, uiOptions));
};
