import { NextFunction, Request, Response, Router } from "express";
import request from "supertest";

import { ApplicationConfig } from "../..";
import { createSecureMiddlewareRouterTestApp } from "../../utils";
import {
  ensureAdminUser,
  ensureAuthUser,
  ensureBasicAuth,
} from "./auth.middleware";
jest.unmock("./auth.middleware.ts");
jest.mock("../../config/env.config.ts");

describe("Middleware Authentication", () => {
  describe("ensureAuthUser", () => {
    const createAuthMiddleware = (isAuth: boolean) => (app: Router) =>
      app.use((req: Request, res: Response, next: NextFunction) => {
        try {
          req.isAuthenticated = () => isAuth;
          ensureAuthUser(req, res, next);
        } catch (e) {
          res.sendStatus(e.statusCode);
        }
      });

    it("should throw a 401 error if user is not authenticated", async () => {
      const isAuth = false;
      const AuthMiddleware = createAuthMiddleware(isAuth);
      const app = createSecureMiddlewareRouterTestApp(AuthMiddleware);
      const res = await request(app).get("/");
      expect(res.status).toBe(401);
    });
  });
  describe("ensureAdminUser", () => {
    const createEnsureAuthMiddleware =
      (isAuth: boolean, blueGroups: string[]) => (app: Router) =>
        app.use((req: Request, res: Response, next: NextFunction) => {
          try {
            req.user = {
              id: "DeMarcus.Kennedy@ibm.com",
              _json: {
                sub: "DeMarcus.Kennedy@ibm.com",
                at_hash: "jNplXUjAPWzOCvx1hCfCbw",
                realmName: "W3IDRealm",
                iss: "url/ism",
                aud: "ZDEwODQ0OGEtNTkwNi00",
                exp: 1595895986,
                iat: 1595888786,
                emailAddress: "DeMarcus.Kennedy@ibm.com",
                lastName: "Kennedy",
                blueGroups,
                cn: "DeMarcus Kennedy",
                dn: "uid=4J8582897,c=us,ou=bluepages,o=ibm.com",
                uid: "4J8582897",
                firstName: "DeMarcus",
                accessToken: "Ahv1aju4KcmI0HQSogNOW4tnZG43qpoZUE2SlbLi",
                refreshToken:
                  "cZCScvHsXhXjUZDd6hfxUaNhMWS16ZdqBAOGto4Md2QlmAnGvo",
              },
            };
            req.isAuthenticated = () => isAuth;
            ensureAdminUser(req, res, () => {
              res.sendStatus(200);
            });
          } catch (e) {
            res.sendStatus(e.statusCode);
          }
        });

    it("should throw a 403 error if user is not an admin", async () => {
      const blueGroups = ["NODE_TEMPLATE"];
      const isAuth = false;
      const AuthMiddleware = createEnsureAuthMiddleware(isAuth, blueGroups);
      const app = createSecureMiddlewareRouterTestApp(AuthMiddleware);
      const res = await request(app).get("/");
      expect(res.status).toBe(403);
    });
    it("should not throw a 403 error if user is an admin", async () => {
      const blueGroups = ["ADMIN"];
      const isAuth = false;
      const AuthMiddleware = createEnsureAuthMiddleware(isAuth, blueGroups);
      const app = createSecureMiddlewareRouterTestApp(AuthMiddleware);
      const res = await request(app).get("/");
      expect(res.status).not.toBe(403);
    });
  });
  describe("ensureBasicAuth", () => {
    const mockAppConfigGet = jest.spyOn(ApplicationConfig, "get");
    const basicAuthConfigFactory = (username?: string, pass?: string) => {
      if (!username && !pass) return {};
      return {
        basicAuth: {
          username,
          pass,
        },
      };
    };
    const getBasicAuth = (username: string, pass: string) => {
      return "Basic " + Buffer.from(`${username}:${pass}`).toString("base64");
    };
    const basicAuthUser = "Jest";
    const basicAuthPass = "Test";
    const createBasicAuthMiddleware = () => (app: Router) => {
      app.use((req: Request, res: Response, next: NextFunction) => {
        try {
          ensureBasicAuth(req, res, next);
          res.sendStatus(200);
        } catch (e) {
          if (e.statusCode) res.sendStatus(e.statusCode);
          res.sendStatus(500); // Lazy mocking here... we anticipate that errors without status code get caught by other middleware, and send 500 to the client
        }
      });
    };

    it("Should return an error if basicAuth is not set", async () => {
      mockAppConfigGet.mockImplementation(() => {
        return basicAuthConfigFactory();
      });
      const basicAuthMiddleware = createBasicAuthMiddleware();
      const app = createSecureMiddlewareRouterTestApp(basicAuthMiddleware);
      const res = await request(app).get("/");
      expect(res.status).toBe(500);
    });
    it("Should return an error if the client does not pass basicAuth credentials", async () => {
      mockAppConfigGet.mockImplementation(() => {
        return basicAuthConfigFactory(basicAuthUser, basicAuthPass);
      });

      const basicAuthMiddleware = createBasicAuthMiddleware();
      const app = createSecureMiddlewareRouterTestApp(basicAuthMiddleware);
      const res = await request(app).get("/");
      expect(res.status).toBe(403);
    });

    it("Should return an error if the client passes in the wrong basicAuth credentials", async () => {
      mockAppConfigGet.mockImplementation(() => {
        return basicAuthConfigFactory(basicAuthUser, basicAuthPass);
      });

      const basicAuthMiddleware = createBasicAuthMiddleware();
      const app = createSecureMiddlewareRouterTestApp(basicAuthMiddleware);
      const res = await request(app)
        .get("/")
        .set("Authorization", getBasicAuth("wrong", "still wrong"));
      expect(res.status).toBe(403);
    });

    it("Should return 200 if passed the right basicAuth credentials", async () => {
      mockAppConfigGet.mockImplementation(() => {
        return basicAuthConfigFactory(basicAuthUser, basicAuthPass);
      });

      const basicAuthMiddleware = createBasicAuthMiddleware();
      const app = createSecureMiddlewareRouterTestApp(basicAuthMiddleware);
      const res = await request(app)
        .get("/")
        .set("Authorization", getBasicAuth(basicAuthUser, basicAuthPass));
      expect(res.status).toBe(200);
    });
  });
});
