import { URLSearchParams } from "url";

import { NextFunction, Request, Response } from "express";
import fetch from "node-fetch";

import { AuthUser, HTTP401Error, HTTP403Error } from "../../common";
import { ApplicationConfig, baseConfig } from "../../config/env.config";
import { isAdministrator } from "../../utils";
import { middlewareLogger } from "./../middleware.api";
import { validateUserWithServiceIdApi } from "./policies.service";
import { AUTH_ROUTES } from "./routes";

type ReqWithSession = Express.User & {
  session: any; // TODO - This used to be Record<string, unknown>, but broke after updating package.json.. Create correct typing - see IESFS-406
};
/**
 * Router-level middleware used to ensure a request has been authenticated.
 *
 * If request has not been authentic, request will be redirected to AUTH_ROUTES.LOGIN.
 *
 * @type {void}
 * @param {Request} req
 * @param {Response} res
 * @param {NextFunction} next
 */

/**
 * Helper function to perform OpenID Connect introspection by making a call
 * to w3id instropection endpoint passing in access token along with the app's
 * client id and client secret. If the response contains a property named 'active'
 * that is set to true, then access token is considered valid.
 *
 * @type {string}
 * @param accesstoken
 */
const introspectAccesstoken = async (accesstoken: string) => {
  const params = new URLSearchParams({
    token: accesstoken,
    client_id: baseConfig.openid.client_id,
    client_secret: baseConfig.openid.client_secret,
  });

  middlewareLogger.info(`Submitting introspection request.`);

  return fetch(baseConfig.openid.introspection_url, {
    method: "post",
    body: params,
  })
    .then((res) => res.json())
    .then((json) => {
      return json;
    });
};

/**
 * The w3id access token is composed of the username and password and
 * is connected to the client_id and client_secret of the app that
 * requested the user to login.
 *
 * The token is parsed out from the authorization header in this expected format:
 * "authorization: Bearer accessToken"
 *
 * @type {string}
 * @param req
 */
const getAccessToken = (req: any): string => {
  try {
    return req.get("authorization").split(" ")[1];
  } catch (e) {
    throw new HTTP401Error();
  }
};

/**
 *
 * Helper function to determine not-authenticated response based on
 * whether the template is being used as a backend-for-frontend or as
 * a pure backend service.
 *
 * @param req
 * @param res
 */
const handleNotAuthenticatedResponse = (req: Request, res: Response) => {
  if (baseConfig.app.serviceType === "BFF") {
    middlewareLogger.info(
      `Request lacks authorization header. Setting session url for redirection.`
    );

    const reqSession = req as ReqWithSession;
    reqSession.session.originalUrl = req.originalUrl;

    middlewareLogger.debug(`Redirecting to ${AUTH_ROUTES.LOGIN} ...`);
    res.redirect(AUTH_ROUTES.LOGIN);
  } else {
    middlewareLogger.info(
      `Request lacks authorization header. Throwing unauthorized error.`
    );
    throw new HTTP401Error();
  }
};

/**
 * Router-level middleware used to ensure a request has been authenticated.
 *
 * If the request has not been authenticated and if the request contains an
 * authorization header, then it implies that this template is being used as
 * a pure backend service and the access token needs to be verified via
 * introspection.
 *
 * Next can be optionally passed as an argument, allowing it to be called outside
 * of this middleware file for manual authentication elsewhere.
 *
 * @type {void}
 * @param {Request} req
 * @param {Response} res
 * @param {NextFunction} next
 */
export const ensureAuthenticated = async (
  req: Request,
  res: Response,
  next?: NextFunction
): Promise<void | boolean> => {
  if (!req.isAuthenticated()) {
    middlewareLogger.info(
      `User failed authentication against protected route: ${req.url}.`
    );
    if (req.get("Authorization")) {
      const accessToken = getAccessToken(req);
      const introspectionResponse = await introspectAccesstoken(accessToken);
      if (introspectionResponse.active) {
        req.user = {
          id: introspectionResponse?.emailAddress,
          _json: introspectionResponse,
        } as AuthUser;
        return next ? next() : true;
      } else {
        middlewareLogger.info(`User's access token is expired or not valid.`);
        throw new HTTP401Error();
      }
    } else {
      handleNotAuthenticatedResponse(req, res);
    }
  } else {
    return next ? next() : true;
  }
};

/**
 * Router-level middleware used to ensure a request has been authenticated.
 *
 * If the request has not been authenticated, request will be redirected to the requests originalUrl
 *
 * @type {void}
 * @param {Request} req
 * @param {Response} res
 * @param {NextFunction} next
 */
export const ensureSessionAuthRedirectUrl = (
  req: Request,
  res: Response,
  next: NextFunction
): void => {
  if (!req.isAuthenticated()) {
    middlewareLogger.info(
      `User failed authentication against protected route: ${req.url}. Setting session url ...`
    );

    const reqSession = req as ReqWithSession;
    reqSession.session.originalUrl = req.originalUrl;
  }
  return next();
};

/**
 * Router-level middleware used to ensure a request has been authenticated.
 *
 * If the request has not been authenticated, HTTP401Error will be thrown.
 *
 * @type {void}
 * @param {Request} req
 * @param {Response} res
 * @param {NextFunction} next
 */
export const ensureAuthUser = (
  req: Request,
  res: Response,
  next: NextFunction
): void => {
  if (!req.isAuthenticated()) {
    middlewareLogger.info(
      `User failed authentication against protected endpoint: ${req.url} `
    );
    throw new HTTP401Error();
  }
  return next();
};

/**
 * Router-level middleware used to validate if a requesting user has the configured admin blue group
 *
 * If the requesting user does not have the configured admin blue group, HTTP403Error will be thrown.
 *
 * @type {void}
 * @param {Request} req
 * @param {Response} res
 * @param {NextFunction} next
 */
export const ensureAdminUser = (
  req: Request,
  res: Response,
  next: NextFunction
): void => {
  const user = req.user as AuthUser; // extends the express user modeling, allowing _json.blueGroups
  if (!isAdministrator(user._json.blueGroups)) {
    middlewareLogger.error(`User not authorized to access this API`);
    throw new HTTP403Error();
  } else {
    next();
  }
};

/**
 * Router-level middleware used to validate a request against the basic auth credentials stored in the server config
 *
 * If credentials are not supplied, or are incorrect, HTTP403Error will be thrown. If a developer doesn't provide basicAuth
 * in the config, an error will get thrown. The desired behavior in this case is to have the server return 500.
 *
 * @type {void}
 * @param {Request} req
 * @param {Response} res
 * @param {NextFunction} next
 */
export const ensureBasicAuth = (
  req: Request,
  res: Response,
  next: NextFunction
): void => {
  // Developer errors - Basic Auth credentials are not stored in the config
  if (
    !ApplicationConfig.get().basicAuth?.username ||
    !ApplicationConfig.get().basicAuth?.pass
  ) {
    middlewareLogger.error(
      "Application config lacks basicAuth credentials to compare against"
    );
    throw new Error("BasicAuth not set");
  }

  const userAuthHeader = req.get("Authorization");
  if (!userAuthHeader) {
    middlewareLogger.info(
      `User failed authentication against ${req.url}: No authorization was provided`
    );
    throw new HTTP403Error("No authorization provided");
  }
  const headersMatch =
    userAuthHeader ===
    "Basic " +
      Buffer.from(
        `${ApplicationConfig.get().basicAuth.username}:${
          ApplicationConfig.get().basicAuth.pass
        }`
      ).toString("base64");
  if (!headersMatch) {
    middlewareLogger.info(
      `User failed authentication against ${req.url}: Invalid credentials provided`
    );
    throw new HTTP403Error("Invalid credentials provided");
  }
  return next();
};

/**
 * Router-level middleware used to ensure a user has accepted the applications global policy
 *
 * If the request does not have a user property, the request will be redirected to AUTH_ROUTES.LOGIN allow the user to login.
 *
 * If the user making the request has not accepted the applications policy, validateUserWithServiceIdApi, the request will be redirected to the req.requestUrl
 *
 * @type {void}
 * @param {Request} req
 * @param {Response} res
 * @param {NextFunction} next
 */
export const requirePolicy = async (
  req: Request,
  res: Response,
  next: NextFunction
): Promise<void> => {
  if (req.user) {
    const config = ApplicationConfig.get();
    if (config.common?.policiesApi) {
      const authUser = req.user as AuthUser;
      const user = await validateUserWithServiceIdApi(authUser.id, config);

      if (user.accept) {
        middlewareLogger.info(
          `${authUser._json.emailAddress} accepted policy!`
        );
        return next();
      }

      middlewareLogger.info("User did not accept policy!");
      middlewareLogger.debug(`Redirecting to ${user.redirectUrl} ...`);
      res.redirect(user.redirectUrl);
    } else {
      middlewareLogger.info("Application has no policy configured");
      return next();
    }
  } else {
    middlewareLogger.info(
      "User is no longer authenticated when attempting to check policy!"
    );
    middlewareLogger.debug(`Redirecting to ${AUTH_ROUTES.LOGIN} ...`);
    res.redirect(AUTH_ROUTES.LOGIN);
  }
};
