import { NextFunction, Request, Response, Router } from "express";
import request from "supertest";

import { baseConfig } from "../../config";
import { createSecureMiddlewareRouterTestApp } from "../../utils";
import * as policyService from "./policies.service";
import { AUTH_ROUTES } from "./routes";

// TODO migrate to use createRouterTestApp()
describe("Middleware Authentication Login", () => {
  const apiUnderTest = AUTH_ROUTES.IS_LOGGED_IN;
  describe("isloggedIn", () => {
    const setPolicyServiceMock = (policyAccepted: boolean) => {
      jest
        .spyOn(policyService, "validateUserWithServiceIdApi")
        .mockResolvedValueOnce({
          accept: policyAccepted,
          redirectUrl: "config.redirectUrl",
        });
    };
    const createMockAuthReqMiddleware =
      (isAuth: boolean, blueGroups: string[] = []) =>
      (app: Router) =>
        app.use((req: Request, res: Response, next: NextFunction) => {
          req.isAuthenticated = () => isAuth;
          req.user = {
            id: "DeMarcus.Kennedy@ibm.com",
            _json: {
              sub: "DeMarcus.Kennedy@ibm.com",
              at_hash: "jNplXUjAPWzOCvx1hCfCbw",
              realmName: "W3IDRealm",
              iss: "url/ism",
              aud: "ZDEwODQ0OGEtNTkwNi00",
              exp: 1595895986,
              iat: 1595888786,
              emailAddress: "DeMarcus.Kennedy@ibm.com",
              lastName: "Kennedy",
              blueGroups,
              cn: "DeMarcus Kennedy",
              dn: "uid=4J8582897,c=us,ou=bluepages,o=ibm.com",
              uid: "4J8582897",
              firstName: "DeMarcus",
              accessToken: "Ahv1aju4KcmI0HQSogNOW4tnZG43qpoZUE2SlbLi",
              refreshToken:
                "cZCScvHsXhXjUZDd6hfxUaNhMWS16ZdqBAOGto4Md2QlmAnGvo",
            },
          };
          next();
        });

    it("should return a 401 if the user is not authenticated", async () => {
      const isAuth = false;
      const isNotAuthenticatedMiddleware = createMockAuthReqMiddleware(isAuth);
      const app = createSecureMiddlewareRouterTestApp(
        isNotAuthenticatedMiddleware
      );

      const res = await request(app).get(apiUnderTest);

      expect(res.status).toBe(401);
    });

    it("should throw a 403 error if the user is authenticated but not authorized", async () => {
      const isAuth = true;
      const isNotAuthenticatedMiddleware = createMockAuthReqMiddleware(isAuth); // no bluegroups -> not authorized
      const app = createSecureMiddlewareRouterTestApp(
        isNotAuthenticatedMiddleware
      );

      const res = await request(app).get(apiUnderTest);

      expect(res.status).toBe(403);
    });

    it("should throw a 451 error if the user is authenticated and authorized but privacy policy is not accepted", async () => {
      const isAuth = true;
      const blueGroups = ["NODE_TEMPLATE"];
      const policyAccepted = false;

      setPolicyServiceMock(policyAccepted);

      const isNotAuthenticatedMiddleware = createMockAuthReqMiddleware(
        isAuth,
        blueGroups
      );
      const app = createSecureMiddlewareRouterTestApp(
        isNotAuthenticatedMiddleware
      );

      const res = await request(app).get(apiUnderTest);

      expect(res.status).toBe(451);
    });

    it("should return the user object if authorized, authenticated, and privacy policy is accepted", async () => {
      const isAuth = true;
      const policyAccepted = true;
      const blueGroups = ["NODE_TEMPLATE"];

      setPolicyServiceMock(policyAccepted);

      const isNotAuthenticatedMiddleware = createMockAuthReqMiddleware(
        isAuth,
        blueGroups
      );
      const app = createSecureMiddlewareRouterTestApp(
        isNotAuthenticatedMiddleware
      );

      const res = await request(app).get(apiUnderTest);
      expect(res.body).toMatchObject({
        user: {
          id: "DeMarcus.Kennedy@ibm.com",
          firstName: "DeMarcus",
          lastName: "Kennedy",
          isAdmin: false,
        },
      });
    });

    it("should return the logged in user object with admin status as true if the user is authenticated, authorized, privacy policy is accepted and is an admin", async () => {
      const isAuth = true;
      const policyAccepted = true;
      const blueGroups = ["ADMIN", "NODE_TEMPLATE"]; // matches .env file used for tests

      setPolicyServiceMock(policyAccepted);

      const isNotAuthenticatedMiddleware = createMockAuthReqMiddleware(
        isAuth,
        blueGroups
      );
      const app = createSecureMiddlewareRouterTestApp(
        isNotAuthenticatedMiddleware
      );

      const res = await request(app).get(apiUnderTest);

      expect(res.body).toMatchObject({
        user: {
          id: "DeMarcus.Kennedy@ibm.com",
          firstName: "DeMarcus",
          lastName: "Kennedy",
          isAdmin: true,
        },
      });
    });

    it("should return the logged in user object with admin as false if user is authenticated, authorized, privacy policy is accepted and not an admin", async () => {
      const isAuth = true;
      const policyAccepted = true;
      const blueGroups = ["notAdmin", "NODE_TEMPLATE"]; // matches .env file used for tests

      setPolicyServiceMock(policyAccepted);

      const isNotAuthenticatedMiddleware = createMockAuthReqMiddleware(
        isAuth,
        blueGroups
      );
      const app = createSecureMiddlewareRouterTestApp(
        isNotAuthenticatedMiddleware
      );

      const res = await request(app).get(apiUnderTest);

      expect(res.body).toMatchObject({
        user: {
          id: "DeMarcus.Kennedy@ibm.com",
          firstName: "DeMarcus",
          lastName: "Kennedy",
          isAdmin: false,
        },
      });
    });
  });

  describe("unauthorizedRedirect", () => {
    it("should redirect to the configured unauthorized redirect url", async () => {
      const app = createSecureMiddlewareRouterTestApp();

      await request(app)
        .get(AUTH_ROUTES.UNAUTH_REDIRECT)
        .expect("Location", baseConfig.app.unauthorizedRedirectUrl);
    });
  });
});
