import { Request, Response, Router } from "express";

import { HTTP401Error, HTTP403Error, HTTP451Error } from "../../common";
import { logMiddlewareStart, middlewareLogger } from "../middleware.api";
import { AuthUser } from "./../../common/types/auth";
import { ApplicationConfig, baseConfig } from "./../../config/env.config";
import { ensureAuthenticated, requirePolicy } from "./auth.middleware";
import { validateUserWithServiceIdApi } from "./policies.service";
import { AUTH_ROUTES } from "./routes";
import { AuthenticatedUser } from "./user.model";

export const handleAuthVerification = (router: Router): void => {
  logMiddlewareStart("Auth Verification");

  // Sanity test api check to make sure its working
  router.get(
    AUTH_ROUTES.TEST_AUTH,
    ensureAuthenticated,
    (req: Request, res: Response) => {
      res.json({ user: req.user });
    }
  );

  router.get(
    AUTH_ROUTES.POLICY_CHECK,
    requirePolicy,
    (req: Request, res: Response) => {
      middlewareLogger.debug(`Redirecting to / ...`);
      res.redirect("/");
    }
  );

  router.get(AUTH_ROUTES.IS_LOGGED_IN, async (req: Request, res: Response) => {
    middlewareLogger.trace("Checking if user is logged in ...");

    if (!req.isAuthenticated()) {
      middlewareLogger.warn("User was not authenticated to be a IBMer");
      throw new HTTP401Error();
    }

    const user = new AuthenticatedUser(req.user as AuthUser);
    middlewareLogger.info(`${user.email} is authenticated to be a IBMer`);

    if (!user.isAuthorized) {
      middlewareLogger.warn(`${user.email} is not authorized to use this app`);
      throw new HTTP403Error();
    }

    middlewareLogger.info(`${user.email} is authorized to use this app`);

    middlewareLogger.trace(
      "Checking if the user has accepted the privacy agreement ..."
    );

    const config = ApplicationConfig.get(); //get updated policies
    if (config.common?.hasOwnProperty("policiesApi")) {
      //dont check policies if not present
      const authUser = await validateUserWithServiceIdApi(user.id, config);

      if (!authUser.accept) {
        middlewareLogger.warn(
          `${user.email} has not accepted the app's privacy agreement`
        );
        throw new HTTP451Error();
      }
      middlewareLogger.info(
        `${user.email} has accepted the app's privacy agreement`
      );
    } else {
      middlewareLogger.info("Application has no policy configured");
    }

    res.json({ user });
  });

  router.get(AUTH_ROUTES.UNAUTH_REDIRECT, (req: Request, res: Response) => {
    middlewareLogger.debug(
      `Redirecting to ${baseConfig.app.unauthorizedRedirectUrl}`
    );
    res.redirect(baseConfig.app.unauthorizedRedirectUrl);
  });

  // TODO investigate why req.baseUrl is not set; see tests, had to create isApiRequest util method
  // router.use(config.app.apiBasePath, verificationRouter);
};
