import { NextFunction, Request, Response, Router } from "express";

import * as ErrorHandler from "../utils/errorHandler.utils";
import { isApiRequest } from "../utils/route.utils";
import { MiddlewareError } from "./../common/types/middleware";
import { logMiddlewareStart } from "./middleware.api";

/**
 * Don’t handle 404 error in its middleware, instead directly propagate it
 * further for a dedicated client errors handler.
 * @param router
 */
const handle404Error = (router: Router): void => {
  logMiddlewareStart("404 Error");

  router.use((req: Request) => {
    if (isApiRequest(req)) {
      ErrorHandler.throwNotFoundError();
    }
  });
};

/**
 * Second middleware in the chain
 * Handles 4xx HTTP errors and if it’s not a case we propagate it down the chain.
 * @param router
 */
const handleClientError = (router: Router): void => {
  logMiddlewareStart("4XX Client Errors");

  router.use(
    (err: MiddlewareError, req: Request, res: Response, next: NextFunction) => {
      if (isApiRequest(req)) {
        ErrorHandler.clientError(err, res, next);
      }
    }
  );
};

/**
 * Our last resort for handling errors must be handled here as a catch all for 5XX errors
 * or uncaughtException handler will be called, and the app node process will exit.
 * Note that even though serverError does not consume next, it must be passed next in order for
 * the middleware function to be called.
 * @param router
 */
const handleServerError = (router: Router): void => {
  logMiddlewareStart("5XX Server Errors");

  router.use(
    (err: MiddlewareError, req: Request, res: Response, next: NextFunction) => {
      ErrorHandler.serverError(err, res, next);
    }
  );
};

export default [handle404Error, handleClientError, handleServerError];
