import { Router } from "express";
import helmet from "helmet";

import { isProduction } from "../config/env.config";
import { logMiddlewareStart } from "./middleware.api";

export const handleHttpGuard = (
  router: Router,
  helmetOptions?: Record<string, unknown>
) => {
  /**
   * Helmet is collection of smaller middleware functions that set security-related HTTP response headers:
   *
   * - csp sets the Content-Security-Policy header to help prevent cross-site scripting attacks and other cross-site injections.
   * - hidePoweredBy removes the X-Powered-By header.
   * - hsts sets Strict-Transport-Security header that enforces secure (HTTP over SSL/TLS) connections to the server.
   * - ieNoOpen sets X-Download-Options for IE8+.
   * - noCache sets Cache-Control and Pragma headers to disable client-side caching.
   * - noSniff sets X-Content-Type-Options to prevent browsers from MIME-sniffing a response away from the declared content-type.
   * - frameguard sets the X-Frame-Options header to provide clickjacking protection.
   */
  if (isProduction) {
    logMiddlewareStart("Http Header Security");
    if (helmetOptions) {
      helmet(helmetOptions);
    } else {
      router.use(
        helmet({
          contentSecurityPolicy: false,
          expectCt: { enforce: false },
          referrerPolicy: { policy: "strict-origin-when-cross-origin" }, // default according to mdn - https://developer.mozilla.org/en-US/docs/Web/HTTP/Headers/Referrer-Policy
        })
      );
    }
  }
};
