import fs from "fs";
import path from "path";

import appRoot from "app-root-path";
import { Express, NextFunction, Request, Response } from "express";

import { CustomExpressMiddleware } from "../../src/utils/app.utils";
import { ApplicationConfig, AuthUser } from "..";
import { CommonServicesProvider } from "../services/common/providers";
import { ensureSessionAuthRedirectUrl } from "./auth/auth.middleware";
import { AuthenticatedUser } from "./auth/user.model";
import { logMiddlewareStart, middlewareLogger } from "./middleware.api";

// Configure this as needed
const RELATIVE_ROOT_BUILD_PATH = "client/build";
const CLIENT_BUILD_PATH = path.join(`${appRoot}`, RELATIVE_ROOT_BUILD_PATH);

export const handleServeStaticContent = (
  app: Express,
  swagger: boolean,
  customServeStaticMiddleware?: CustomExpressMiddleware[]
): void => {
  logMiddlewareStart(`Serve Static Content`);
  middlewareLogger.debug(`Serving from: ${CLIENT_BUILD_PATH}`);

  // If swagger is enabled, we need to serve the redirect page for swagger when the redirect url is hit
  if (swagger) {
    app.get(
      "/oauth2-redirect.html*",
      ensureSessionAuthRedirectUrl,
      (req, res) => {
        res.set(
          "Content-Security-Policy",
          "default-src https: 'unsafe-eval' 'unsafe-inline'; object-src 'none'"
        );
        res.sendFile(
          `${appRoot}/node_modules/swagger-ui-dist/oauth2-redirect.html`
        );
      }
    );
  }

  const CustomExpressMiddleware = customServeStaticMiddleware
    ? customServeStaticMiddleware
    : []; // if undefined or null assigned to empty array so spread below works

  app.get(
    "*",
    handleClientFileRequest,
    ensureSessionAuthRedirectUrl,
    handleMaintenanceMode,
    ...CustomExpressMiddleware,
    async (req: Request, res: Response) => {
      // Proxy all other requests to the frontend
      // Default with session auth guard to allow client to redirect back after login
      // in IES the default deployment of static files is through node which is why we don't ignore the proxy in production
      middlewareLogger.debug(`Maintenance mode not enabled for this request`);
      middlewareLogger.info(`Serving index.html for user`);
      res.sendFile(`${CLIENT_BUILD_PATH}/index.html`);
    }
  );
};

const serveIndexOrMaintenancePage = async (req: Request, res: Response) => {
  const isUnderMaintenance = await fetchMaintenanceStatus();
  if (isUnderMaintenance) {
    serveMaintenancePage(req, res);
    return;
  }
  serveIndexPage(req, res);
};

const fetchMaintenanceStatus = async () => {
  const config = ApplicationConfig.get();
  const provider = new CommonServicesProvider({});
  try {
    const { underMaintenance } = await provider.getServiceById(
      config.common?.serviceId
    );
    middlewareLogger.debug(
      `Service data retrieved, maintenance mode is set to ${underMaintenance}`
    );
    return underMaintenance;
  } catch (e) {
    middlewareLogger.error(`Failed to fetch service data: ${e}`);
    return false;
  }
};

const serveMaintenancePage = (req: Request, res: Response) => {
  middlewareLogger.info(`Serving maintenance.html to client`);
  const path = `${CLIENT_BUILD_PATH}/maintenance.html`;
  fs.stat(path, (err, stat) => {
    if (err) return res.status(404).send();
    if (stat.isDirectory()) return res.status(404).send();
    return res.sendFile(path);
  });
};

const serveIndexPage = (req: Request, res: Response) => {
  middlewareLogger.info(`Serving index.html to client`);
  const path = `${CLIENT_BUILD_PATH}/index.html`;
  fs.stat(path, (err, stat) => {
    if (err) return res.status(404).send();
    if (stat.isDirectory()) return res.status(404).send();
    return res.sendFile(path);
  });
};

/**
 * Serves a file client requested if it exists, if not continues
 */
const handleClientFileRequest = async (
  req: Request,
  res: Response,
  next: NextFunction
) => {
  const path = `${CLIENT_BUILD_PATH}${req.path}`;
  middlewareLogger.debug(`Servicing file request at ${path}`);
  fs.stat(path, (err, stat) => {
    if (err) return next();
    if (stat.isDirectory()) return next();
    return res.sendFile(path);
  });
};

/**
 * If server is running in maintenance mode, will call service to determine application status
 * Admin users see the index.html page, all other users see either index.html or maintenance.html
 * depending on maintenance status
 */
const handleMaintenanceMode = async (
  req: Request,
  res: Response,
  next: NextFunction
) => {
  const config = ApplicationConfig.get();
  if (config.app.enableMaintenanceMode) {
    if (!req.isAuthenticated()) {
      middlewareLogger.debug(
        `User is not authenticated, determining index or maintenance page to serve`
      );
      await serveIndexOrMaintenancePage(req, res);
      return;
    }

    const user = new AuthenticatedUser(req.user as AuthUser);
    const isAdmin = user.isAdmin;

    if (isAdmin) {
      middlewareLogger.debug(
        "Admin user: will serve index.html regardless of maintenance status"
      );
      await serveIndexPage(req, res);
      return;
    }

    middlewareLogger.debug(
      "Standard user: determining index of maintenance page to serve"
    );
    await serveIndexOrMaintenancePage(req, res);
    return;
  }
  next();
};
