import http, { Server } from "http";
import https from "https";
import { AddressInfo } from "net";

import { AppManager } from "./app";
import { logger } from "./common";
import { AppServerConfig, baseConfig } from "./config/env.config";
import { configureAppHealth } from "./health";

export interface AppServer {
  start(): void;
}

export class ApplicationServer implements AppServer {
  private static readonly LOGGER = logger.child("Server");
  private readonly server: Server;
  private config: AppServerConfig;

  constructor(appManager: AppManager, config: AppServerConfig) {
    ApplicationServer.LOGGER.info("Creating Server".toUpperCase());
    const app = appManager.app();

    this.config = config;
    this.server =
      this.config.https && baseConfig.app.useHttps
        ? https.createServer(this.config.https, app)
        : http.createServer(app);

    configureAppHealth(this.server);
  }

  start(): void {
    ApplicationServer.LOGGER.log("Starting Server".toUpperCase());

    this.server.listen(this.config.port, () => {
      const addressInfo = this.server.address() as AddressInfo;
      const address =
        addressInfo.address === "::" ? "localhost" : addressInfo.address;
      const protocol =
        this.config.https && baseConfig.app.useHttps ? "https" : "http";

      const ServerRunningMessage = `Started on ${protocol}://${address}:${this.config.port}/`;

      ApplicationServer.LOGGER.info(ServerRunningMessage);
    });

    process.on(
      "uncaughtException",
      logger.exitHandler("uncaughtException", () => process.exit(1))
    );

    // TODO migrate to logger.exitHandler if typescript error can be resolved
    process.on("unhandledRejection", (e) => {
      logger.log(e);
      process.exit(1);
    });
  }
}
