import { NextFunction, Request, Response } from "express";
import { buildCheckFunction, ValidationChain } from "express-validator";
import { CREATED, OK } from "http-status-codes";

import {
  AbstractRestController,
  BusinessViolationCode,
  CrudController,
  logger,
  LoggerApi,
  Model,
  RestResponse,
} from "./../../common";
import { User } from "./user.model";
import { UserService } from "./user.service";

interface RequestWithUser extends Request {
  user: User | undefined;
}

export class UserController
  extends AbstractRestController<User>
  implements CrudController
{
  constructor(private usersService: UserService) {
    super();
  }

  private static readonly LOGGER: LoggerApi = logger.child("user");

  getAll(req: Request, res: Response): Promise<any> {
    UserController.LOGGER.debug("Retrieving all users");

    return this.usersService
      .getAll()
      .then((users: User[]) => this.respond(res, users));
  }

  get(req: Request, res: Response): Promise<any> {
    return Promise.resolve(this.respond(res, (req as RequestWithUser)?.user));
  }

  create(req: Request, res: Response): Promise<any> {
    const user = new User(req.body);
    this.validateModel(user);

    if (user.email === "ies@ibm.com") {
      this.throwBusinessViolation(
        BusinessViolationCode.BLACKLISTED_EMAIL,
        "This email is blacklisted."
      );
    }

    return this.usersService.create(user).then((createdUser: User) => {
      return this.respond(res, createdUser, CREATED);
    });
  }

  update(req: Request, res: Response): Promise<any> {
    const userToUpdate: User = (req as RequestWithUser).user as User;
    userToUpdate.set(req.body);
    this.validateModel(userToUpdate);

    return this.usersService
      .update(userToUpdate.id || "", userToUpdate) // The or statement here gets around a TS error, the right-hand side should never execute
      .then((updatedUser: User) => {
        return this.respond(res, updatedUser);
      });
  }

  delete(req: Request, res: Response): Promise<any> {
    return this.usersService
      .delete((req as RequestWithUser).user?.id || "")
      .then(() => {
        return this.respondNoContent(res);
      });
  }

  async query(req: Request, res: Response): Promise<any> {
    const nodeOrg = req.query.nodeOrg as string;
    const info = await this.usersService.repoInfo(nodeOrg);
    // TODO: create a model instead of bypass warning with cast
    const response = new RestResponse(info as unknown as Model);
    return res.status(OK).send(response);
  }

  resolveUser(
    req: Request,
    res: Response,
    next: NextFunction,
    param: string
  ): Promise<User | void> {
    return this.usersService.get(param).then((user: User | null) => {
      if (user) {
        (req as RequestWithUser).user = user;
      }
      next();
    });
  }

  queryValidator(): ValidationChain {
    const checkQueryParams = buildCheckFunction(["query"]);
    return checkQueryParams("nodeOrg").exists();
  }

  requestValidator(): ValidationChain {
    const checkQueryParams = buildCheckFunction(["params"]);
    return checkQueryParams("userId").exists();
  }
}
