import { AbstractRestRouter } from "../../common";
import { UserController } from "./user.controller";
import { User } from "./user.model";
import { UserService } from "./user.service";

// Swagger is commented out here to prevent it from being read when exported
// /**
//  * @swagger
//  * tags:
//  *   name: Users
//  *   description: User management
//  */
export class UserRouter extends AbstractRestRouter<User, UserController> {
  protected basePath = "/users";

  private controller: UserController;

  constructor(usersService: UserService) {
    super();
    this.controller = new UserController(usersService);
    this.initRoutes();
  }

  initRoutes() {
    const c = this.controller;

    this.router.param("userId", this.handleRouteParam(c, c.resolveUser));

    // /**
    //  * @swagger
    //  * path:
    //  *  /users/:
    //  *    get:
    //  *      summary: Get all users
    //  *      tags: [Users]
    //  *      responses:
    //  *        "200":
    //  *          description: An array of users
    //  *          content:
    //  *            application/json:
    //  *              schema:
    //  *                type: array
    //  *                items:
    //  *                  $ref: '#/components/schemas/User'
    //  */
    this.router.get("/", this.handleRoute(c, c.getAll));

    // /**
    //  * @swagger
    //  * path:
    //  *  /users/:
    //  *    post:
    //  *      summary: Create a new user
    //  *      tags: [Users]
    //  *      requestBody:
    //  *        required: true
    //  *        content:
    //  *          application/json:
    //  *            schema:
    //  *              $ref: '#/components/schemas/User'
    //  *      responses:
    //  *        "200":
    //  *          description: A user schema
    //  *          content:
    //  *            application/json:
    //  *              schema:
    //  *                $ref: '#/components/schemas/User'
    //  */
    this.router.post("/", this.handleRoute(c, c.create));
    this.router.all("/", this.handleRoute(c, c.throwMethodNotAllowedError));

    this.router.get(
      "/repoOrg",
      [c.queryValidator()],
      this.handleRoute(c, c.query)
    );

    // /**
    //  * @swagger
    //  * path:
    //  *  /users/{userId}:
    //  *    get:
    //  *      summary: Get a user by id
    //  *      tags: [Users]
    //  *      parameters:
    //  *        - in: path
    //  *          name: userId
    //  *          schema:
    //  *            type: string
    //  *          required: true
    //  *          description: Id of the user
    //  *      responses:
    //  *        "200":
    //  *          description: An users object
    //  *          content:
    //  *            application/json:
    //  *              schema:
    //  *                $ref: '#/components/schemas/User'
    //  */
    this.router.get(
      "/:userId",
      [c.requestValidator()],
      this.handleRoute(c, c.get)
    );
    this.router.delete(
      "/:userId",
      [c.requestValidator()],
      this.handleRoute(c, c.delete)
    );
    this.router.patch(
      "/:userId",
      [c.requestValidator()],
      this.handleRoute(c, c.update)
    );
    this.router.all(
      "/:userId",
      this.handleRoute(c, c.throwMethodNotAllowedError)
    );
  }
}
