import { CrudService } from "../../common";
import GithubProvider from "./providers/GithubProvider";
import { User } from "./user.model";

export type UserCreationParams = Pick<User, "email" | "name" | "phoneNumbers">;

export class UserService implements CrudService<User> {
  private userMap: Map<string, User>;
  private maxUserId = 0;
  private repositoryProvider: GithubProvider = new GithubProvider();

  constructor() {
    const defaultUser = new User({ id: 1 });
    defaultUser.email = "jane@doe.com";
    defaultUser.name = "Jane Doe";
    defaultUser.status = "Happy";
    defaultUser.phoneNumbers = [];

    this.userMap = this.populateUserMap([defaultUser]);

    const userIds = Array.from(this.userMap.keys()).map((val) =>
      parseInt(val, 10)
    );

    this.maxUserId = Math.max(...userIds);
  }

  async repoInfo(org: string): Promise<string> {
    return this.repositoryProvider.findOrg(org);
  }

  getAll(): Promise<User[]> {
    const users = Array.from(this.userMap.values());
    return Promise.resolve(users);
  }

  get(id: string): Promise<User | null> {
    const user = this.userMap.get(id);

    if (user == null) {
      return Promise.resolve(null);
    }

    return Promise.resolve(user);
  }

  create(user: User): Promise<User> {
    const newUserId = ++this.maxUserId;
    user.id = newUserId.toString();

    this.userMap.set(user.id, user);
    return Promise.resolve(user);
  }

  update(id: string, user: User): Promise<User> {
    this.userMap.set(id, user);
    return Promise.resolve(user);
  }

  delete(id: string): Promise<void> {
    this.userMap.delete(id);
    return Promise.resolve();
  }

  private populateUserMap(users: User[]): Map<string, User> {
    const usersMap = new Map<string, User>();

    users.forEach((user, i) => usersMap.set(user.id || `${i}`, user));

    return usersMap;
  }
}
