import { ObjectId } from "mongodb";
import { MongoMemoryServer } from "mongodb-memory-server";
import mongoose, { Connection } from "mongoose";

class MongoMemoryClient {
  private mongod!: MongoMemoryServer;
  private connection!: Connection;

  private connectToDB = async () => {
    this.mongod = await MongoMemoryServer.create();
    const uri = this.mongod.getUri();

    await mongoose.connect(uri);
    this.connection = mongoose.connection;
  };

  private async setConnection() {
    await this.connectToDB();
    return this.connection;
  }

  /**
   * replaces 'id' with '_id' prop with values of 'new ObjectId(data.id);'
   *
   * @param {Object}data document to be inserted
   * @returns {Object} document to be inserted with mongo id
   */
  private addSeededMongoId(data: any) {
    const { id, ...dataWithoutId } = data;
    dataWithoutId._id = new ObjectId(data.id);
    data = dataWithoutId;
    return data;
  }

  async getConnection() {
    if (!this.connection) {
      return await this.setConnection();
    } else {
      return this.connection;
    }
  }

  /**
   * adds data to the specified collection
   *
   * @param {string} collectionName
   * @param {object} data data to be seeded to the specified collection
   */
  async seedCollection(collectionName: string, data: any) {
    if (!Array.isArray(data)) data = [data]; //turn into array so loop below handles all cases
    data = data.reduce((acc: any[], cur: any) => {
      if (cur.id) {
        cur = this.addSeededMongoId(cur);
      }
      acc.push(cur);
      return acc;
    }, []);
    await this.setConnection();
    const collection = await this.connection.createCollection(collectionName);
    await collection.insertMany(data);
  }

  /**
   * deletes all of the collections created
   */
  async clearDatabase() {
    const collections = mongoose.connection.collections;
    const deletePromises = [];

    for (const key in collections) {
      const collection = collections[key];
      deletePromises.push(collection.deleteMany({}));
    }

    return await Promise.all(deletePromises);
  }

  /**
   * closes connection with db
   */
  async closeDatabase() {
    await mongoose.connection.dropDatabase();
    await mongoose.connection.close();
    await this.mongod.stop();
  }
}

export const mongoMemoryClient = new MongoMemoryClient();
