import { Headers, Response } from "node-fetch";

import { ExternalApiError } from "./../../common/rest/errors/externalApi.error";
import { HttpApiJsonRequester } from "./apiRequester.utils";

describe("API Requester", () => {
  const body = { name: "Bob Robertson", email: "bob@gmail.com" };
  const defaultOptions = { compress: false };
  const defaultContentTypeRequestHeader = {
    "Content-Type": "application/json",
  };
  const externalDefaultOptions = {
    follow: 10,
    headers: HttpApiJsonRequester.oauth2Header("token"), // this should conflict and merge with private static defaults
  };

  const apiRequester = new HttpApiJsonRequester(
    "test",
    externalDefaultOptions,
    false
  );

  describe("fetchJson()", () => {
    const url = "/api/users/0";

    const headers = new Headers([["Content-Type", "application/json"]]);

    const successResponse: Partial<Response> = {
      ok: true,
      status: 200,
      headers,
      json: () => Promise.resolve(body),
    };

    const errorResponse: Partial<Response> = {
      ok: false,
      status: 500,
      statusText: "Internal server error",
    };

    const emptyResponse: Partial<Response> = {
      ok: true,
      status: 204,
      headers,
    };

    function mockFetchSuccess() {
      return jest
        .spyOn(apiRequester, "fetch")
        .mockImplementation(jest.fn().mockResolvedValue(successResponse))
        .mockName("fetch");
    }

    function mockFetchError() {
      return jest
        .spyOn(apiRequester, "fetch")
        .mockImplementation(jest.fn().mockResolvedValue(errorResponse))
        .mockName("fetch");
    }

    function mockFetchEmpty() {
      return jest
        .spyOn(apiRequester, "fetch")
        .mockImplementation(jest.fn().mockResolvedValue(emptyResponse))
        .mockName("fetch");
    }

    afterEach(() => {
      jest.restoreAllMocks();
    });

    test("calls fetch with the specified url and options", async () => {
      const options = { compress: true };
      const spy = mockFetchSuccess();
      await apiRequester.fetchJson(url, options);

      const allOptions = {
        ...externalDefaultOptions,
        ...options,
        headers: {
          ...externalDefaultOptions.headers,
          ...defaultContentTypeRequestHeader,
        },
      };
      expect(spy).toHaveBeenCalledWith(url, allOptions);
    });

    test("uses the default (and configured) fetch options if none are specified", async () => {
      const spy = mockFetchSuccess();
      await apiRequester.fetchJson(url);

      const allOptions = {
        ...defaultOptions,
        ...externalDefaultOptions,
        headers: {
          ...externalDefaultOptions.headers,
          ...defaultContentTypeRequestHeader,
        },
        method: undefined,
      };
      expect(spy).toHaveBeenCalledWith(url, allOptions);
    });

    test("resolves to the JSON-parsed response body of a success response", () => {
      mockFetchSuccess();
      return expect(
        apiRequester.fetchJson(url, {
          headers: { "Content-Type": "application/json" },
        })
      ).resolves.toBe(body);
    });

    test("resolves to a raw response of a successful response that has no content (204 status)", () => {
      mockFetchEmpty();
      return expect(
        apiRequester.fetchJson(url, {
          headers: { "Content-Type": "application/json" },
        })
      ).resolves.toBe(undefined);
    });

    test("rejects with an error including the status text for an error response", () => {
      mockFetchError();
      const expectedError = apiRequester.error(
        errorResponse.status as number,
        errorResponse.statusText
      );

      return expect(apiRequester.fetchJson(url)).rejects.toThrow(expectedError);
    });

    test("includes the status code and the full response in the error", () => {
      mockFetchError();

      const expectedError = apiRequester.error(
        errorResponse.status as number,
        errorResponse.statusText
      );

      return apiRequester.fetchJson(url).catch((err) => {
        expect(err).toStrictEqual(expectedError);
      });
    });

    test("should be able to use a api error response transformer", () => {
      const apiName = "test";

      const apiErrorResponseTransformer = async (
        res: Response
      ): Promise<ExternalApiError> => {
        const resp = await res.json();

        const apiError = ExternalApiError.factory(apiName);

        return apiError(res.status, resp);
      };

      const API_REQUESTER = new HttpApiJsonRequester(
        apiName,
        externalDefaultOptions,
        false,
        apiErrorResponseTransformer
      );

      const errorBody = { ...errorResponse, code: "a code" };

      const customApiErrorResponse = {
        status: errorResponse.status,
        json: () => Promise.resolve(errorBody),
      };

      jest
        .spyOn(API_REQUESTER, "fetch")
        .mockImplementation(jest.fn().mockResolvedValue(customApiErrorResponse))
        .mockName("fetch");

      const expectedError = API_REQUESTER.error(
        customApiErrorResponse.status as number,
        errorBody
      );

      return API_REQUESTER.fetchJson(url).catch((err) => {
        expect(err).toStrictEqual(expectedError);
      });
    });

    test("should be able to use a api error response transformer provided in fetchJSON arguments", () => {
      const apiName = "test";

      const apiErrorResponseTransformer = async (
        res: Response
      ): Promise<ExternalApiError> => {
        const resp = await res.json();

        const apiError = ExternalApiError.factory(apiName);

        return apiError(res.status, resp);
      };

      const errorBody = { ...errorResponse, code: "a code" };

      const customApiErrorResponse = {
        status: errorResponse.status,
        json: () => Promise.resolve(errorBody),
      };

      jest
        .spyOn(apiRequester, "fetch")
        .mockImplementation(jest.fn().mockResolvedValue(customApiErrorResponse))
        .mockName("fetch");

      const expectedError = apiRequester.error(
        customApiErrorResponse.status as number,
        errorBody
      );

      return apiRequester
        .fetchJson(url, {}, apiErrorResponseTransformer)
        .catch((err) => {
          expect(err).toStrictEqual(expectedError);
        });
    });
    test("rejects with the fetch error if the request fails altogether", () => {
      const error = new Error("Oops!");
      jest
        .spyOn(apiRequester, "fetch")
        .mockImplementation(() => Promise.reject(error))
        .mockName("fetch");
      return expect(apiRequester.fetchJson(url)).rejects.toThrow(error);
    });
  });

  describe("fetchJsonWithoutDefaultContentType()", () => {
    const url = "/api/users/0";

    const headers = new Headers([["Content-Type", "application/json"]]);

    const successResponse: Partial<Response> = {
      ok: true,
      status: 200,
      headers,
      json: () => Promise.resolve(body),
    };

    const errorResponse: Partial<Response> = {
      ok: false,
      status: 500,
      statusText: "Internal server error",
    };

    function mockFetchSuccess() {
      return jest
        .spyOn(apiRequester, "fetch")
        .mockImplementation(jest.fn().mockResolvedValue(successResponse))
        .mockName("fetch");
    }

    function mockFetchError() {
      return jest
        .spyOn(apiRequester, "fetch")
        .mockImplementation(jest.fn().mockResolvedValue(errorResponse))
        .mockName("fetch");
    }

    afterEach(() => {
      jest.restoreAllMocks();
    });

    test("calls fetch with the specified url and options", async () => {
      const options = { compress: true };
      const spy = mockFetchSuccess();
      await apiRequester.fetchJsonWithoutDefaultContentType(url, options);

      const allOptions = {
        ...externalDefaultOptions,
        ...options,
        headers: {
          ...externalDefaultOptions.headers,
        },
      };
      expect(spy).toHaveBeenCalledWith(url, allOptions);
    });

    test("uses the default (and configured) fetch options if none are specified", async () => {
      const spy = mockFetchSuccess();
      await apiRequester.fetchJsonWithoutDefaultContentType(url);

      const allOptions = {
        ...defaultOptions,
        ...externalDefaultOptions,
        headers: {
          ...externalDefaultOptions.headers,
        },
        method: undefined,
      };
      expect(spy).toHaveBeenCalledWith(url, allOptions);
    });

    test("resolves to the JSON-parsed response body of a success response", () => {
      mockFetchSuccess();
      return expect(
        apiRequester.fetchJsonWithoutDefaultContentType(url, {
          headers: { "Content-Type": "application/json" },
        })
      ).resolves.toBe(body);
    });

    test("rejects with an error including the status text for an error response", () => {
      mockFetchError();
      const expectedError = apiRequester.error(
        errorResponse.status as number,
        errorResponse.statusText
      );

      return expect(
        apiRequester.fetchJsonWithoutDefaultContentType(url)
      ).rejects.toThrow(expectedError);
    });

    test("includes the status code and the full response in the error", () => {
      mockFetchError();

      const expectedError = apiRequester.error(
        errorResponse.status as number,
        errorResponse.statusText
      );

      return apiRequester
        .fetchJsonWithoutDefaultContentType(url)
        .catch((err) => {
          expect(err).toStrictEqual(expectedError);
        });
    });

    test("should be able to use a api error response transformer", () => {
      const apiName = "test";

      const apiErrorResponseTransformer = async (
        res: Response
      ): Promise<ExternalApiError> => {
        const resp = await res.json();

        const apiError = ExternalApiError.factory(apiName);

        return apiError(res.status, resp);
      };

      const API_REQUESTER = new HttpApiJsonRequester(
        apiName,
        externalDefaultOptions,
        false,
        apiErrorResponseTransformer
      );

      const errorBody = { ...errorResponse, code: "a code" };

      const customApiErrorResponse = {
        status: errorResponse.status,
        json: () => Promise.resolve(errorBody),
      };

      jest
        .spyOn(API_REQUESTER, "fetch")
        .mockImplementation(jest.fn().mockResolvedValue(customApiErrorResponse))
        .mockName("fetch");

      const expectedError = API_REQUESTER.error(
        customApiErrorResponse.status as number,
        errorBody
      );

      return API_REQUESTER.fetchJson(url).catch((err) => {
        expect(err).toStrictEqual(expectedError);
      });
    });

    test("should be able to use a api error response transformer provided in fetchJSON arguments", () => {
      const apiName = "test";

      const apiErrorResponseTransformer = async (
        res: Response
      ): Promise<ExternalApiError> => {
        const resp = await res.json();

        const apiError = ExternalApiError.factory(apiName);

        return apiError(res.status, resp);
      };

      const errorBody = { ...errorResponse, code: "a code" };

      const customApiErrorResponse = {
        status: errorResponse.status,
        json: () => Promise.resolve(errorBody),
      };

      jest
        .spyOn(apiRequester, "fetch")
        .mockImplementation(jest.fn().mockResolvedValue(customApiErrorResponse))
        .mockName("fetch");

      const expectedError = apiRequester.error(
        customApiErrorResponse.status as number,
        errorBody
      );

      return apiRequester
        .fetchJsonWithoutDefaultContentType(
          url,
          {},
          apiErrorResponseTransformer
        )
        .catch((err) => {
          expect(err).toStrictEqual(expectedError);
        });
    });
    test("rejects with the fetch error if the request fails altogether", () => {
      const error = new Error("Oops!");
      jest
        .spyOn(apiRequester, "fetch")
        .mockImplementation(() => Promise.reject(error))
        .mockName("fetch");
      return expect(
        apiRequester.fetchJsonWithoutDefaultContentType(url)
      ).rejects.toThrow(error);
    });
  });

  describe("fetchText()", () => {
    const url = "/api/users/0";
    const responseText = "Bob";
    const headers = new Headers([["Content-Type", "application/text"]]);
    const successResponse: Partial<Response> = {
      ok: true,
      status: 200,
      headers,
    };

    function mockFetchSuccess() {
      return jest
        .spyOn(apiRequester, "fetch")
        .mockImplementation(jest.fn().mockResolvedValue(successResponse))
        .mockName("fetch");
    }

    afterEach(() => {
      jest.restoreAllMocks();
    });

    test.skip("resolves to the text response body of a success response", () => {
      mockFetchSuccess();
      return expect(apiRequester.fetchJson(url)).resolves.toBe("");
    });
  });

  describe("postJson()", () => {
    test("returns an options object with the expected method and headers properties", () => {
      const options = HttpApiJsonRequester.postJson(body);
      expect(options).toHaveProperty("method", "POST");
      expect(options).toHaveProperty("headers", {
        "Content-Type": "application/json",
      });
    });

    test("converts the specified body object to JSON and includes it in the options", () => {
      const bodyJson = '{"name":"Bob Robertson","email":"bob@gmail.com"}';
      expect(HttpApiJsonRequester.postJson(body)).toHaveProperty(
        "body",
        bodyJson
      );
    });

    test("includes the default options if they are not overridden", () => {
      expect(HttpApiJsonRequester.postJson(body)).toMatchObject(defaultOptions);
    });

    test("overrides and adds options as specified", () => {
      const options = HttpApiJsonRequester.postJson(body, {
        compress: true,
        size: 1000,
      });
      expect(options).toHaveProperty("compress", true);
      expect(options).toHaveProperty("size", 1000);
    });
  });

  describe("deleteJson()", () => {
    test("returns an options object with the expected method and headers properties", () => {
      const options = HttpApiJsonRequester.deleteJson(body);
      expect(options).toHaveProperty("method", "DELETE");
      expect(options).toHaveProperty("headers", {
        "Content-Type": "application/json",
      });
    });

    test("converts the specified body object to JSON and includes it in the options", () => {
      const bodyJson = '{"name":"Bob Robertson","email":"bob@gmail.com"}';
      expect(HttpApiJsonRequester.deleteJson(body)).toHaveProperty(
        "body",
        bodyJson
      );
    });

    test("includes the default options if they are not overridden", () => {
      expect(HttpApiJsonRequester.deleteJson(body)).toMatchObject(
        defaultOptions
      );
    });

    test("overrides and adds options as specified", () => {
      const options = HttpApiJsonRequester.deleteJson(body, {
        compress: true,
        size: 1000,
      });
      expect(options).toHaveProperty("compress", true);
      expect(options).toHaveProperty("size", 1000);
    });
  });

  describe("putJson()", () => {
    test("returns an options object with the expected method and headers properties", () => {
      const options = HttpApiJsonRequester.putJson(body);
      expect(options).toHaveProperty("method", "PUT");
      expect(options).toHaveProperty("headers", {
        "Content-Type": "application/json",
      });
    });

    test("converts the specified body object to JSON and includes it in the options", () => {
      const bodyJson = '{"name":"Bob Robertson","email":"bob@gmail.com"}';
      expect(HttpApiJsonRequester.putJson(body)).toHaveProperty(
        "body",
        bodyJson
      );
    });

    test("includes the default options if they are not overridden", () => {
      expect(HttpApiJsonRequester.putJson(body)).toMatchObject(defaultOptions);
    });

    test("overrides and adds options as specified", () => {
      const options = HttpApiJsonRequester.putJson(body, {
        compress: true,
        size: 1000,
      });
      expect(options).toHaveProperty("compress", true);
      expect(options).toHaveProperty("size", 1000);
    });
  });

  describe("patchJson()", () => {
    test("returns an options object with the expected method and headers properties", () => {
      const options = HttpApiJsonRequester.patchJson(body);
      expect(options).toHaveProperty("method", "PATCH");
      expect(options).toHaveProperty("headers", {
        "Content-Type": "application/json",
      });
    });

    test("converts the specified body object to JSON and includes it in the options", () => {
      const bodyJson = '{"name":"Bob Robertson","email":"bob@gmail.com"}';
      expect(HttpApiJsonRequester.patchJson(body)).toHaveProperty(
        "body",
        bodyJson
      );
    });

    test("includes the default options if they are not overridden", () => {
      expect(HttpApiJsonRequester.patchJson(body)).toMatchObject(
        defaultOptions
      );
    });

    test("overrides and adds options as specified", () => {
      const options = HttpApiJsonRequester.patchJson(body, {
        compress: true,
        size: 1000,
      });
      expect(options).toHaveProperty("compress", true);
      expect(options).toHaveProperty("size", 1000);
    });
  });

  describe("oauth2Header()", () => {
    test("should create an aoth 2.0 bearer token header", () => {
      const token = "a token";

      const oathHeader = HttpApiJsonRequester.oauth2Header(token);

      expect(oathHeader).toMatchObject({
        Authorization: `Bearer ${token}`,
      });
    });
  });

  describe("basicAuth()", () => {
    test("should create a base 64 encoded string from two variables", () => {
      const authUsername = "demarcus.kennedy@ibm.com";
      const authPassword = "abc123DEF456";
      const basicAuth = HttpApiJsonRequester.basicAuth(
        authUsername,
        authPassword
      );

      expect(basicAuth).toMatchObject({
        Authorization:
          "Basic " +
          Buffer.from(`${authUsername}:${authPassword}`).toString("base64"),
      });
    });
  });
});
