import fetch, { RequestInfo, RequestInit, Response } from "node-fetch";
import { Omit } from "utility-types";

import { logger, LoggerApi } from "../../common/logger";
import { ExternalApiError } from "../../common/rest/errors/externalApi.error";
import { traceFetchRequests } from "./apiTracer.utils";

export interface APIRequester {
  fetchJson: (
    url: RequestInfo,
    options?: RequestInit,
    errorResponseTransformer?: ApiErrorResponseTransformer
  ) => Promise<any>;

  fetchJsonWithoutDefaultContentType: (
    url: RequestInfo,
    options?: RequestInit
  ) => Promise<any>;
}

type RequestOptions = Omit<RequestInit, "method">;
type ApiErrorResponseTransformer = (res: Response) => Promise<ExternalApiError>;

/**
 *  Utility that wraps fetch to simplify its use with JSON-based APIs.
 *  TODO consider making a HttpApiJsonRequesterBuilder
 *  TODO consider making a RequestBuilder from the static methods
 */
export class HttpApiJsonRequester implements APIRequester {
  private static readonly LOGGER: LoggerApi = logger.child(
    "HttpApiJsonRequester"
  );
  private static readonly defaultOptions: RequestInit = { compress: false };
  private static readonly headers = { "Content-Type": "application/json" };

  private externalDefaultInstanceOptions: RequestInit;
  private apiErrorResponseTransformer?: ApiErrorResponseTransformer;
  private request: typeof fetch;

  get fetch(): typeof fetch {
    return this.request;
  }

  set fetch(f: typeof fetch) {
    this.request = f;
  }

  private apiError: (
    status: number,
    message: Record<string, unknown> | string | undefined
  ) => ExternalApiError;

  constructor(
    api: string,
    defaultOptions: RequestInit = {},
    enableTracing = true,
    apiErrorResponseTransformer?: ApiErrorResponseTransformer
  ) {
    this.externalDefaultInstanceOptions = defaultOptions;
    this.request = enableTracing ? traceFetchRequests(api) : fetch;
    this.apiErrorResponseTransformer = apiErrorResponseTransformer;
    this.apiError = ExternalApiError.factory(api);
  }

  /**
   * Apply ad-hoc scoped (hardcoded deep) merge for internal default props only
   * If default props become deeply nested consider using
   * a deep merge lib such as lodash _.merge
   * @param method
   * @param options
   */
  private static mergeProps(
    method: string,
    options: RequestInit = {},
    useDefaultContentType = true
  ) {
    return {
      ...HttpApiJsonRequester.defaultOptions,
      ...options,
      method, // Force method to not be overriden by options
      headers: {
        ...HttpApiJsonRequester.defaultOptions.headers,
        ...(useDefaultContentType ? HttpApiJsonRequester.headers : {}),
        ...options.headers,
      },
    };
  }

  static getJson(options: RequestOptions = {}): Record<string, unknown> {
    return this.mergeProps("GET", options);
  }

  static deleteJson(
    body = {},
    options: RequestOptions = {}
  ): Record<string, unknown> {
    return this.mergeProps("DELETE", {
      ...options,
      body: JSON.stringify(body),
    });
  }

  static postJson(
    body = {},
    options: RequestOptions = {}
  ): Record<string, unknown> {
    return this.mergeProps("POST", {
      ...options,
      body: JSON.stringify(body),
    });
  }

  static putJson(
    body = {},
    options: RequestOptions = {}
  ): Record<string, unknown> {
    return this.mergeProps("PUT", {
      ...options,
      body: JSON.stringify(body),
    });
  }

  static patchJson(
    body = {},
    options: RequestOptions = {}
  ): Record<string, unknown> {
    return this.mergeProps("PATCH", {
      ...options,
      body: JSON.stringify(body),
    });
  }

  static oauth2Header(accessToken: string): { Authorization: string } {
    return {
      Authorization: `Bearer ${accessToken}`,
    };
  }

  static basicAuth(
    username: string,
    password: string
  ): { Authorization: string } {
    return {
      Authorization:
        "Basic " + Buffer.from(`${username}:${password}`).toString("base64"),
    };
  }

  get error(): (
    status: number,
    message: Record<string, unknown> | undefined | string
  ) => ExternalApiError {
    return this.apiError;
  }

  /**
   * Enhances fetch() as follows:
   * 1. It retrieves the body of a success response, parses it, and resolves to the resulting object.
   * 2. It rejects on receiving an error response, with an error that captures key reponse details.
   * 3. It applies default options if none are specified.
   * @param url the request url
   * @param options the fetch() configuration options
   * @param customApiErrorResponseTransformer optional Error Transformer that can be used in liue of throwing the default this.apiError
   */
  fetchJson(
    url: RequestInfo,
    options: RequestInit = {},
    customApiErrorResponseTransformer?: ApiErrorResponseTransformer
  ): Promise<string | Record<string, unknown>> {
    const instanceOptions: RequestInit = {
      ...this.externalDefaultInstanceOptions,
      ...options,
      headers: {
        ...this.externalDefaultInstanceOptions.headers,
        ...options.headers,
      },
    };

    const allOptions: RequestInit = HttpApiJsonRequester.mergeProps(
      instanceOptions.method as string,
      instanceOptions
    );

    const toResponseMessage = async (response: Response) => {
      const contentType = response.headers?.get("content-type");
      const status = response.status;

      if (!contentType || status === 204) {
        return;
      }

      if (contentType.indexOf("application/json") !== -1) {
        return response.json();
      } else {
        return response.text();
      }
    };

    return this.fetch(url, allOptions).then(async (response: Response) => {
      const errorTransformer = customApiErrorResponseTransformer
        ? customApiErrorResponseTransformer
        : this.apiErrorResponseTransformer;
      if (!response.ok) {
        HttpApiJsonRequester.LOGGER.info(await toResponseMessage(response));
        if (errorTransformer) {
          HttpApiJsonRequester.LOGGER.debug(
            "Using the custom http error response transformer"
          );
          throw await errorTransformer(response);
        }
        throw this.apiError(response.status, response.statusText);
      }
      return toResponseMessage(response);
    });
  }

  /**
   * Enhances fetchJson() as follows:
   * 1. It sends requests without the HttpApiJsonRequester.headers property
   * @param url the request url
   * @param options the fetch() configuration options
   * @param customApiErrorResponseTransformer optional Error Transformer that can be used in liue of throwing the default this.apiError
   */
  fetchJsonWithoutDefaultContentType(
    url: RequestInfo,
    options: RequestInit = {},
    customApiErrorResponseTransformer?: ApiErrorResponseTransformer
  ): Promise<string | Record<string, unknown>> {
    const instanceOptions: RequestInit = {
      ...this.externalDefaultInstanceOptions,
      ...options,
      headers: {
        ...this.externalDefaultInstanceOptions.headers,
        ...options.headers,
      },
    };

    const allOptions: RequestInit = HttpApiJsonRequester.mergeProps(
      instanceOptions.method as string,
      instanceOptions,
      false
    );

    const toResponseMessage = async (response: Response) => {
      const contentType = response.headers?.get("content-type");

      if (!contentType) {
        return;
      }

      if (contentType.indexOf("application/json") !== -1) {
        return response.json();
      } else {
        return response.text();
      }
    };

    return this.fetch(url, allOptions).then(async (response: Response) => {
      const errorTransformer = customApiErrorResponseTransformer
        ? customApiErrorResponseTransformer
        : this.apiErrorResponseTransformer;
      if (!response.ok) {
        HttpApiJsonRequester.LOGGER.info(await toResponseMessage(response));
        if (errorTransformer) {
          HttpApiJsonRequester.LOGGER.debug(
            "Using the custom http error response transformer"
          );
          throw await errorTransformer(response);
        }
        throw this.apiError(response.status, response.statusText);
      }
      return toResponseMessage(response);
    });
  }
}
