import { Router } from "express";

import { baseConfig } from "../config";
import coreMiddleware, { errorMiddleware } from "../middleware";
import { handleHttpGuard } from "../middleware/httpGuard.middleware";
import { routes } from "../services";
import { AppFactory } from "../utils/app.utils";
import { ApiRouterFactory } from "./route.utils";

// We have to import and mock the function here because creating a mock from jest.fn() does not assign the correct name
// e.g. saying const handleHttpGuard = jest.fn() will give a value of "mockConstructor" when calling handleHttpGuard.name
// For testing, we need handleHttpGuard.name to be handleHttpGuard
jest.mock("../middleware/httpGuard.middleware.ts");

function mockFunction<T extends (...args: any[]) => any>(
  fn: T
): jest.MockedFunction<T> {
  return fn as jest.MockedFunction<T>;
}

const mockedHandleHttpGuard = mockFunction(handleHttpGuard);

describe("App Utils", () => {
  describe("ApiRouterFactory Class", () => {
    describe("#makeRouter", () => {
      test.skip("should create express application", () => {
        const router: Router = ApiRouterFactory.makeRouter(routes);

        const app = AppFactory.createExpressApp(
          baseConfig.app,
          router,
          coreMiddleware,
          errorMiddleware
        );
        expect(app).toBeTruthy();
      });
    });
  });
  describe("preApiRouterMiddleware", () => {
    test("Should call httpGuard middleware with custom options if httpGuard middleware options are provided to the middleware config object", () => {
      const router: Router = ApiRouterFactory.makeRouter(routes);

      const helmetConfig = {
        contentSecurityPolicy: false,
        expectCt: { enforce: false },
        referrerPolicy: false,
      };

      const customCoreMiddleware = [mockedHandleHttpGuard];

      AppFactory.createExpressApp(
        baseConfig.app,
        router,
        customCoreMiddleware,
        errorMiddleware,
        { preApiRouterMiddleware: { handleHttpGuard: helmetConfig } }
      );

      expect(mockedHandleHttpGuard).toHaveBeenCalledTimes(1);
      expect(mockedHandleHttpGuard).toHaveBeenCalledWith(
        expect.anything(),
        helmetConfig
      );
    });

    test("Should call httpGuard middleware with no options if httpGuard middleware options are not provided to the middleware config object", () => {
      const router: Router = ApiRouterFactory.makeRouter(routes);

      const customCoreMiddleware = [mockedHandleHttpGuard];

      AppFactory.createExpressApp(
        baseConfig.app,
        router,
        customCoreMiddleware,
        errorMiddleware
      );

      expect(mockedHandleHttpGuard).toHaveBeenCalledTimes(1);
      expect(mockedHandleHttpGuard).toHaveBeenCalledWith(expect.anything());
    });
  });
});
