import express, {
  Express,
  NextFunction,
  Request,
  Response,
  Router,
} from "express";
import { Options } from "swagger-jsdoc";

import { logger } from "../common";
import { AppConfig, GraphQLConfig } from "../config/env.config";
import { handleLogging, handleServeStaticContent } from "../middleware";
import { handleAPIDocs } from "../middleware/apiDocs.middleware";
import { handleGraphQL } from "../middleware/graphQL.middleware";
import { MiddlewareConfigurer } from "../middleware/middleware.api";

export type CustomExpressMiddleware = (
  req: Request,
  res: Response,
  next: NextFunction
) => void;

export interface MiddlewareOptions {
  graphQLConfig?: GraphQLConfig;
  swagger?: {
    config: Options;
    uiOptions?: Record<string, unknown>;
  };
  preApiRouterMiddleware?: Record<string, unknown>;
  customServeStaticMiddleware?: CustomExpressMiddleware[];
}

export class AppFactory {
  private static readonly LOGGER = logger.child("App Factory");
  static createExpressApp(
    appConfig: AppConfig,
    apiRouter: Router,
    preApiRouterMiddleware: MiddlewareConfigurer[],
    postApiRouterMiddleware: MiddlewareConfigurer[],
    middlewareOptions: MiddlewareOptions = { preApiRouterMiddleware: {} }
  ): Express {
    const app: Express = express();

    if (appConfig.enableHttpRequestLogging) {
      handleLogging(app);
    }

    if (appConfig.swagger) {
      handleAPIDocs(
        app,
        middlewareOptions.swagger?.config,
        middlewareOptions.swagger?.uiOptions
      );
    }

    if (preApiRouterMiddleware != null) {
      this.LOGGER.info(
        `Loading Pre-Router Middleware [${preApiRouterMiddleware.length}]`.toUpperCase()
      );
      preApiRouterMiddleware.forEach((middleware) => {
        const customOptions = middlewareOptions.preApiRouterMiddleware
          ? middlewareOptions.preApiRouterMiddleware[middleware.name]
          : null;
        if (customOptions) {
          middleware(app, customOptions);
        } else {
          middleware(app);
        }
      });
    }

    this.LOGGER.info("Mounting Routes".toUpperCase());
    app.use(appConfig.apiBasePath, apiRouter);

    const numPostRouterMiddleware =
      postApiRouterMiddleware.length + Number(appConfig.serveStatic);

    this.LOGGER.info(
      `Loading Post-Router Middleware [${numPostRouterMiddleware}]`.toUpperCase()
    );

    // NOTE: This must be applied after mounting routes but before post router middleware since it hangs!
    if (middlewareOptions.graphQLConfig) {
      handleGraphQL(app, middlewareOptions.graphQLConfig);
    }

    if (appConfig.serveStatic) {
      handleServeStaticContent(
        app,
        appConfig.swagger ? true : false,
        middlewareOptions?.customServeStaticMiddleware
      );
    }

    if (postApiRouterMiddleware != null) {
      postApiRouterMiddleware.forEach((middleware) => middleware(apiRouter));
    }

    return app;
  }
}
