import {
  ExternalApiError,
  HTTP404Error,
  HTTPClientError,
} from "../common/rest/errors";
import { clientError, throwNotFoundError } from "./errorHandler.utils";
import { mockResponse } from "./test.utils";

describe("Error Handler Utils", () => {
  describe("#throwNotFoundError", () => {
    it("should throw HTTP404Error", () => {
      expect(throwNotFoundError).toThrow(HTTP404Error);
    });
  });

  describe("#clientError", () => {
    const testStatusCode = 499;
    const testErrMessage = "Test Error";

    class TestClientError extends HTTPClientError {
      readonly statusCode = testStatusCode;

      constructor(message: string | Record<string, unknown> = testErrMessage) {
        super(message);
      }
    }

    // tslint:disable: max-classes-per-file
    class TestMiddlewareError extends Error {
      readonly statusCode = 100;
    }

    const newTestExternalApiError = (statusCode: number = testStatusCode) =>
      new ExternalApiError("test api", statusCode, testErrMessage);

    it("should send the error if its an HTTPClientError then terminate", () => {
      const mockResp = mockResponse();
      const mockNext = jest.fn();
      const httpClientError = new TestClientError();

      clientError(httpClientError, mockResp as any, mockNext);

      expect(mockResp.status).toBeCalledWith(httpClientError.statusCode);
      expect(mockResp.send).toBeCalledWith(httpClientError);
      expect(mockNext).not.toBeCalled();
    });
    it("should send the error if its a 4XX ExternalApiError then terminate", () => {
      const mockResp = mockResponse();
      const mockNext = jest.fn();
      const externalApiError = newTestExternalApiError();

      clientError(externalApiError, mockResp as any, mockNext);

      expect(mockResp.status).toBeCalledWith(externalApiError.statusCode);
      expect(mockResp.send).toBeCalledWith(externalApiError);
      expect(mockNext).not.toBeCalled();
    });

    it("should not send the error if its a ExternalApiError with no 4XX statusCode and then continue error middleware chain", () => {
      const mockResp = mockResponse();
      const mockNext = jest.fn();
      const non4xxStatusCode = 300;
      const externalApiError = newTestExternalApiError(non4xxStatusCode);

      clientError(externalApiError, mockResp as any, mockNext);

      expect(mockResp.send).not.toBeCalled();
      expect(mockNext).toBeCalled();
    });

    it("should not send the error if its not a ExternalApiError or HttpClientError and then continue error middleware chain", () => {
      const mockResp = mockResponse();
      const mockNext = jest.fn();
      const testMiddlewareError = new TestMiddlewareError();

      clientError(testMiddlewareError, mockResp as any, mockNext);

      expect(mockResp.send).not.toBeCalled();
      expect(mockNext).toBeCalled();
    });
  });
});
