import { NextFunction, Response } from "express";
import { getStatusText, INTERNAL_SERVER_ERROR } from "http-status-codes";

import {
  ExternalApiError,
  HTTP404Error,
  HTTPClientError,
  logger,
} from "../common";
import { isProduction } from "../config/env.config";
import { MiddlewareError } from "./../common/types/middleware";

const CLIENT_ERROR_REGEX = RegExp("4[0-9][0-9]");

export const throwNotFoundError = (): void => {
  throw new HTTP404Error("Method not found");
};

export const isClientError = (err: MiddlewareError): boolean =>
  err instanceof HTTPClientError ||
  (err instanceof ExternalApiError &&
    CLIENT_ERROR_REGEX.test(err.statusCode.toString()));

export const clientError = (
  err: MiddlewareError,
  res: Response,
  next: NextFunction
): void => {
  if (isClientError(err)) {
    logger.error(err);
    res.status(err.statusCode).send(err);
  } else {
    next(err);
  }
};

export const serverError = (
  err: MiddlewareError,
  res: Response,
  next: NextFunction
): void => {
  logger.error(err);
  // We want to send up errors from error transformer functions in our providers, so send up the errors if it has a message property
  if (err.message) {
    res.status(err.statusCode).send(err);
  } else if (isProduction) {
    res
      .status(INTERNAL_SERVER_ERROR)
      .send(getStatusText(INTERNAL_SERVER_ERROR));
  } else {
    res.status(INTERNAL_SERVER_ERROR).send(err.stack);
  }
};
