import express, { Application } from "express";

import { RestRouter } from "../common/rest/restRouter.api";
import { AppConfig, baseConfig } from "../config";
import { handleLogin } from "../middleware/auth/login.middleware";
import { handleAuthVerification } from "../middleware/auth/verification.middleware";
import { handleBodyRequestParsing } from "../middleware/common.middleware";
import errorHandlerMiddleware from "../middleware/errorHandler.middleware";
import { MiddlewareConfigurer } from "./../middleware/middleware.api";
import { handleSessions } from "./../middleware/session.middleware";
import { AppFactory } from "./app.utils";
import { ApiRouterFactory } from "./route.utils";

export const createRouterTestPath = (
  restRouter: RestRouter,
  testPath = ""
): string =>
  `${baseConfig.app.apiBasePath}${restRouter.getBasePath()}${testPath}`;
export const createRouterTestApp = (restRouter: RestRouter): Application => {
  const preApiMiddleware = [handleBodyRequestParsing];
  const postApiMiddleware = errorHandlerMiddleware;

  const appConfig: AppConfig = {
    apiBasePath: baseConfig.app.apiBasePath,
  };

  const router = ApiRouterFactory.makeRouter([restRouter]);

  const app = AppFactory.createExpressApp(
    appConfig,
    router,
    preApiMiddleware,
    postApiMiddleware,
    { preApiRouterMiddleware: {} } // Indicates no custom middleware options
  );

  return app;
};

// TODO phase out or implement differently
export const createSecureMiddlewareRouterTestApp = (
  ...middleware: MiddlewareConfigurer[]
): Application => {
  const app = express();

  handleBodyRequestParsing(app);

  // custom test middleware
  middleware.forEach((m) => m(app));

  handleAuthVerification(app);

  errorHandlerMiddleware.forEach((m) => m(app));

  return app;
};

export const createSecureRouterTestApp = (
  restRouter: RestRouter
): Application => {
  const preApiMiddleware = [
    handleBodyRequestParsing,
    handleSessions,
    handleLogin,
  ];
  const postApiMiddleware = errorHandlerMiddleware;
  const appConfig: AppConfig = {
    apiBasePath: baseConfig.app.apiBasePath,
  };

  const router = ApiRouterFactory.makeRouter([restRouter]);

  const app = AppFactory.createExpressApp(
    appConfig,
    router,
    preApiMiddleware,
    postApiMiddleware,
    { preApiRouterMiddleware: {} }
  );

  return app;
};

type MockResponse = {
  status: jest.MockedFunction<any>;
  json: jest.MockedFunction<any>;
  send: jest.MockedFunction<any>;
};

export const mockResponse = (): MockResponse => {
  const res: MockResponse = {
    status: "",
    json: "",
    send: "",
  };

  res.status = jest.fn().mockReturnValue(res);
  res.json = jest.fn().mockReturnValue(res);
  res.send = jest.fn().mockReturnValue(res);

  return res;
};

/**
 * Sets up common jest app needs globally
 * Note: "any" type must be used in ensureAuthenticated in order
 *       to do the object assignments
 */
export const setupGlobals = (): void => {
  // WHY DOESN'T THIS WORK??
  // jest
  //   .spyOn(LoginMiddleware, "ensureAuthenticated")
  //   .mockImplementation((req, rest, next) => {
  //     next();
  //   });
  jest.mock("../middleware/auth/auth.middleware", () => ({
    ensureAuthenticated: (req: any, rest: any, next: any) => {
      req.user = {
        _json: {
          accessToken: "blah",
        },
      };
      next();
    },
  }));

  global.afterEach(() => jest.clearAllMocks());
};
